'use strict';

const express = require('express');
const http = require('http');
const path = require('path');
const socketIO = require('socket.io');
const app = express();
const server = http.Server(app);
const io = socketIO(server);
const yargs = require('yargs').argv;
const PORT = process.env.PORT || 7000;
const { performance } = require('perf_hooks');

const GAME_FPS = 200; //100; //30
//const GR_FPS = 100; //30
const mSpeedC = 30 / GAME_FPS * 1.2;
const FIELD_SC = 4;
const FIELD_WIDTH = 1000*FIELD_SC, FIELD_HEIGHT = 1000*FIELD_SC;
const START_TALL = 0; //1000 * 4;
var mFieldGridSize = 240;
var mFieldGridSizeH = 180;
var mBulletSize = 6;
var mBulletSpeed = mFieldGridSize * 1.0;
var mBulletLength = mBulletSpeed*1.0;
var mPlayerSpeed = mFieldGridSize/30 * mSpeedC; //8
var mPlayerTurnSpeed = 0.05 * mSpeedC;
var mPlayerWidth = 40;
var mPlayerTall = 90;
var mWallSize = mFieldGridSize;
var mWallSizeH = mFieldGridSizeH;
var mWallThick = 20 * 1.0;
var mFloorThick = 20 * 1.0;
var mWallVisualThick =  5 * 1.0;
var mFloorVisualThick = 5 * 1.0;
var mShootCycle = [Math.floor(6/mSpeedC), Math.floor(30/mSpeedC), Math.floor(60/mSpeedC)]; //AR,SG,S
var mMagagineSize = [30,8,1];
var mJumpHight = mWallSizeH * 0.7;
var mJumpFrame = 24 / mSpeedC; //30
var mBandageSize = 20;
var mFallSpeed = 18 * mSpeedC; //15

function mFloorGS(v_){ //floor to grid size
    return Math.floor(v_/mFieldGridSize) * mFieldGridSize;
}
function mFloorGSH(v_){ //floor to grid size H
    return Math.floor(v_/mFieldGridSizeH) * mFieldGridSizeH;
}

class GameObject{
    constructor(obj={}){
        this.id = Math.floor(Math.random()*1000000000);
        this.x = obj.x;
        this.y = obj.y;
        this.z = obj.z; //
        this.width = obj.width;
        this.height = obj.height;
        this.tall = obj.tall;
        this.Lx = obj.width;
        this.Ly = obj.height;
        this.Lz = obj.tall;
        this.angle = obj.angle;
        this.angle2 = obj.angle2;
        this.isMoving = 0;
        this.isSquat = 0;
        this.isEditMode = 0;
    }
    move(distance){
        const oldX = this.x, oldY = this.y;
        this.isMoving = 1;

        this.x += distance * Math.cos(this.angle);
        this.y += distance * Math.sin(this.angle);
    }
    side(distance){
        const oldX = this.x, oldY = this.y;
        this.isMoving = 1;
        
        this.x += distance * Math.cos(this.angle+3.1415/2);
        this.y += distance * Math.sin(this.angle+3.1415/2);        
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, width: this.width, height: this.height, tall:this.tall, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, angle: this.angle, angle2: this.angle2, isMoving: this.isMoving}; //, isEditMode: this.isEditMode};
    }
}


class Player extends GameObject{
    constructor(obj={}){
        super(obj);
        this.socketId = obj.socketId;
        this.nickname = obj.nickname;
        this.width = 80;
        this.height = 80;
        this.tall = mPlayerTall; //original
        this.eyeTall = mPlayerTall * 0.8;
        this.health = this.maxHealth = 100;
        this.shield = this.maxShield = 100;
        this.bullets = {};
        this.weapon = 1;
        this.point = 0;
        this.movement = {};
        this.shootCount = [0,0,0];
        this.isShoot = 0;
        this.isSquat = 0;
        this.timeCount = 0;
        this.shootTime = 0;
        this.firingTime = 0;
        this.giveDamage = 0;
        this.giveDamageTime = 0;
        this.isBot = 0;
        this.isGrounded = 1;
        this.isJump = 0;
        this.jumpTime = 0;
        this.jumpStartZ = 0;
        this.bandages = 0;
        this.potions = 0;
        this.isEmote = 0;
        this.mode = 0; // 0:falling, 1:shoot, 2:build
        this.buildTemp = null;
        this.buildType = 0; //0:wall, 1:floor
        this.inStorm = 0;
        this.oldX = 0;
        this.oldY = 0;
        this.oldZ = 0;
        this.velocity = [0,0,0];
        this.nowTouboBuild = 0;
        this.isSelectingGrid = 0;
        this.editBuildID = -1;
        this.lastDamegePlayer = -1;
        this.cameraPosition = [0, 0, 0];
        this.isSkinLoaded = 0;
        this.order = -1;

        this.z = 0;//START_TALL;
        //this.angle = Math.random() * 2 * Math.PI;
        this.angle2 = 0;
        this.x = 100;
        this.y = 100;
        this.angle = 0;
        this.z = START_TALL;

        this.velocity = [0, 0, 0];
        this.cameraPosition = [-150, -30, this.tall*0.8];

    }
    getItem(obj){
        var judge = false;
        var v = Math.sqrt( (this.x-obj.x)*(this.x-obj.x)
                          +(this.y-obj.y)*(this.y-obj.y)
                          +(this.z-obj.z)*(this.z-obj.z)  );
        if(v<=mPlayerWidth){
           judge = true; 
        }//
        return judge;
    }
    getHight(){
        var v = this.tall * (1.0 -0.3*this.isSquat);
        return v;
    }
    shoot(){

        if(this.mode != 1 ){
            return;
        }

        const bullet = new Bullet({
            x: this.x + this.width/2 * Math.cos(this.angle),
            y: this.y + this.width/2 * Math.sin(this.angle),
            z: this.z , //+ (this.tall -this.tall/2*this.isSquat )*0.8
            angle: this.angle,
            angle: this.angle2,
            player: this,
            damageV: 20,
        });
        //console.log('bullet angle:'+bullet.angle+', '+bullet.angle2);

        bullet.x = this.x - this.cameraPosition[1] * Math.sin(this.angle);
        bullet.y = this.y + this.cameraPosition[1] * Math.cos(this.angle);
        bullet.z = this.z + this.cameraPosition[2];

        bullet.dx = Math.cos(-this.angle2) * Math.cos(this.angle);
        bullet.dy = Math.cos(-this.angle2) * Math.sin(this.angle);
        bullet.dz = Math.sin(-this.angle2);

        //console.log('bullet d:'+bullet.dx+', '+bullet.dy+', '+bullet.dz);
        
        if(this.weapon==2){
            bullet.damageV = 75;
            bullet.height = mBulletSize*5;
            bullet.tall = mBulletSize*5;
        }
        else if(this.weapon==3){
            bullet.damageV = 120;
        }

        //bullet.move(this.width/2);
        //bullet.side(this.width/2*Math.cos(this.angle));
        this.bullets[bullet.id] = bullet;
        bullets[bullet.id] = bullet;

        //io.to(this.socketId).emit('shoot', this.weapon);
        //console.log('shoot from;'+this.socketId);
        io.sockets.emit('shoot',[this.socketId, this.weapon]);
    }
    funcWall(){

        var angle_ = Math.floor( (this.angle+3.1415/4) / (3.1415/2)) * 3.1415/2;
        var angle2_ = 0;
        var x_ = mFloorGS(this.x);
        var y_ = mFloorGS(this.y);
        var z_ = mFloorGSH(this.z + 10);
        var dz_ =  this.z - z_;
        if( (dz_>mFieldGridSizeH/2) && (this.angle2<-3.1415/6) ){
            z_ += mFieldGridSizeH;
        }

        // var Lx_ = mWallThick;
        // var Ly_ = mWallSize;
        // var Lz_ = mWallSizeH;
        var L_ = [mWallThick, mWallSize, mWallSizeH];

        let M = [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ];
        var offset_ = [0, 0, 0];

        if( (angle_==0) || 
            (angle_==3.1415*2) )
        {
            if( this.x>=x_ ){
                x_ += mFieldGridSize;
            }
        }
        else if(angle_==3.1415)
        {
            if( this.x>=x_+mFieldGridSize ){
                x_ += mFieldGridSize;
            }
        }//
        else if(angle_==3.1415/2)
        {
            if( this.y>=y_ ){
                y_ += mFieldGridSize;
            }    
            M = [
                [0, -1, 0],
                [1, 0, 0],
                [0, 0, 1]
            ];
            offset_ = [mWallSize, 0, 0];
        }
        else if(angle_==3.1415/2*3)
        {
            if( this.y>=y_+mFieldGridSize ){
                y_ += mFieldGridSize;
            }   
            M = [
                [0, -1, 0],
                [1, 0, 0],
                [0, 0, 1]
            ]; 
            offset_ = [mWallSize, 0, 0];
        }

        L_ = mTranslatePhase(L_, M, [0,0,0]);
        L_ = mAbsVec(L_);

        //let wall = mCreatePieceOfWall(x_,y_,z_,L_[0],L_[1],L_[2]);
        let wall = mCreateWall(x_,y_,z_,L_[0],L_[1],L_[2]);
        wall.M = M;
        wall.offset = offset_;
        wall.angle = angle_;
        wall.angle2 = angle2_;

        //console.log('nvec:'+wall.nvec);

        if( (wall.angle==3.1415/2) ||
            (wall.angle==3.1415/2*3)  )
        {
            let pvec_ = wall.pvec;
            let nvec_ = wall.nvec;
            let index_ =[2,3,0,1];
            for(var i=0;i<4; i++){
                for(var j=0;j<9; j++){
                    wall.pvec[i*9+j] = pvec_[index_[i]*9+j];
                }//
                for(var j=0;j<3; j++){
                    wall.nvec[i*3+j] = nvec_[index_[i]*3+j];
                }//
            }//

            //console.log('nvec:'+wall.nvec);
            //console.log('');    
        }//


        var dw =  4*2;
        var dh =  3*2;
        wall.gridSize[0] = mWallVisualThick; //mWallThick;
        wall.gridSize[1] = (mWallSize - dw*2)/3;
        wall.gridSize[2] = (mWallSizeH - dh*2)/3;
        
        wall.gridL0 = mWallVisualThick; //mWallThick;
        wall.gridL1 = (mWallSize - dw*2)/3;
        wall.gridL2 = (mWallSizeH - dh*2)/3;

        let originVec = Array(9*3).fill(0);

        for(var i=0;i<3; i++){
            for(var j=0;j<3; j++){
                originVec[(i*3+j)*3+0] = 0;
                originVec[(i*3+j)*3+1] = mWallSize - (j+1)*wall.gridSize[1] - j*dw;
                originVec[(i*3+j)*3+2] = mWallSizeH - (i+1)*wall.gridSize[2] - i*dh;
            }//     
        } //


        wall.gridSize = mTranslatePhase(wall.gridSize, wall.M, [0,0,0]);
        let d = [wall.x + wall.offset[0], wall.y + wall.offset[1], wall.z + wall.offset[2]];
        wall.gridOrigin = mTranslatePhase(originVec, wall.M, d);

        for(var i=0;i<3;i++){
            if(wall.gridSize[i]<0){
                wall.gridSize[i] = Math.abs(wall.gridSize[i]);
                for(var k=0;k<9;k++){
                    wall.gridOrigin[3*k+i] -= wall.gridSize[i];
                }
            }
        }


        return wall;
    }
    buildWall(){

        this.mode = 2;
        this.buildType = 0;
        var wall = this.funcWall();
        var j = mCheckUniqueBuild(wall,0);
        if(j){
            this.buildTemp = this.funcWall();
            if(this.nowTouboBuild==1){
                this.doBuild();
            }//
        }//
        else{
            this.buildTemp = null;
        }

    }
    funcFloor()
    {

        var angle_ = this.angle; // + 3.1415/2;
        angle_ = Math.floor( (angle_+3.1415/4) / (3.1415/2)) * 3.1415/2;
        var angle2_ = this.angle2;

        var x_, y_, z_;

        x_ = mFloorGS(this.x);
        y_ = mFloorGS(this.y);
        z_ = mFloorGSH(this.z + mWallSizeH*0.2); //mFallSpeed*2);

        //var dz_ =  this.z - mFloorGSH(this.z);
        //console.log('dz_:'+dz_+', angle2:'+this.angle2);

        if( (angle_==0) || (angle_==3.1415*2) ){
            if( this.x>=x_+mFieldGridSize/2 ){
                x_ += mFieldGridSize;
            }
        }
        else if(angle_==3.1415){
            if( this.x<=x_+mFieldGridSize/2 ){
                x_ -= mFieldGridSize;
            }
        }
        else if(angle_==3.1415/2){
            if( this.y>=y_+mFieldGridSize/2 ){
                y_ += mFieldGridSize;
            }
        }
        else if(angle_==3.1415/2*3){
            if( this.y<=y_+mFieldGridSize/2 ){
                y_ -= mFieldGridSize;
            }
        }

        if( (angle2_<-3.1415/4) || (angle2_>3.1415/4) ){
            x_ = mFloorGS(this.x);
            y_ = mFloorGS(this.y);
        }
        if( angle2_<-3.1415/6 ){
            z_ += mFieldGridSizeH;
        }

        let floor = mCreateFloor(x_, y_, z_, mWallSize, mWallSize, mFloorVisualThick);

        let M = [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ];
        floor.M = M;

        var dw =  20;
        var Lgx = (mWallSize - dw)/2;
        var Lgy = (mWallSize - dw)/2;
        floor.gridSize[0] = (mWallSize - dw)/2;
        floor.gridSize[1] = (mWallSize - dw)/2;
        floor.gridSize[2] = mFloorVisualThick; //mFloorThick;

        floor.gridL0 = floor.gridSize[0];
        floor.gridL1 = floor.gridSize[1];
        floor.gridL2 = floor.gridSize[2];
        
        let originVec = Array(4*3).fill(0);
        let gridP = [];
        let gridF = [];

        for(var i=0;i<2; i++){
            for(var j=0;j<2; j++){
                var x,y,z;
                x = j*floor.gridSize[0] + j*dw;
                y = mWallSize - (i+1)*floor.gridSize[1] - i*dw;
                z = 0; 

                x += floor.x;
                y += floor.y;
                z += floor.z;

                originVec[(i*2+j)*3+0] = x;
                originVec[(i*2+j)*3+1] = y;
                originVec[(i*2+j)*3+2] = z;

                gridP[(i*2+j)*4+0] = [x,y,z];
                gridP[(i*2+j)*4+1] = [x+Lgx,y,z];
                gridP[(i*2+j)*4+2] = [x+Lgx,y+Lgy,z];
                gridP[(i*2+j)*4+3] = [x,y+Lgy,z];

                var k = (i*2+j)*4;
                gridF[(i*2+j)*4+0] = [k+0,k+1,k+2];
                gridF[(i*2+j)*4+1] = [k+0,k+2,k+1];
                gridF[(i*2+j)*4+2] = [k+0,k+2,k+3];
                gridF[(i*2+j)*4+3] = [k+0,k+3,k+2];
            }//     
        } //
        //console.log('gridP size:'+ gridP.length );

        floor.gridOrigin = originVec;
        floor.gridP = gridP;
        floor.gridF = gridF;

        return floor;
    }
    buildFloor(){
       
        this.mode = 2;
        this.buildType = 1;
        var floor = this.funcFloor();
        var j = mCheckUniqueBuild(floor,1);
        if(j){
            this.buildTemp = this.funcFloor();
            if(this.nowTouboBuild==1){
                this.doBuild();
            }//
        }//
        else{
            this.buildTemp = null;
        }
    }
    funcSlope()
    { 
        var angle_ = this.angle; // + 3.1415/2;
        angle_ = Math.floor( (angle_+3.1415/4) / (3.1415/2)) * 3.1415/2;
        angle_ = angle_%(3.1415*2);
 
        // player geo center
        var x_ = this.x;// + this.width/2 * Math.cos(this.angle);
        var y_ = this.y;// + this.width/2 * Math.sin(this.angle);
        x_ = mFloorGS(x_);   //Math.floor(x_/mFieldGridSize)*mFieldGridSize;
        y_ = mFloorGS(y_);   //Math.floor(y_/mFieldGridSize)*mFieldGridSize;

        var z_ = mFloorGSH(this.z + mWallSizeH*0.2) + mFieldGridSizeH/2;
        var dz_ =  this.z - mFloorGSH(this.z);
        //console.log('dz_:'+dz_+', angle2:'+this.angle2);
        if( (dz_>mFieldGridSizeH/2) && (this.angle2<-3.1415/6) ){
            z_ += mFieldGridSizeH;
        }

        x_ += mFieldGridSize * Math.cos(angle_);
        y_ += mFieldGridSize * Math.sin(angle_);

        if( (this.angle2>3.1415/4) ){
            x_ = mFloorGS(this.x);
            y_ = mFloorGS(this.y);
        }

        let slope = mCreateSlope(x_,y_,z_,mWallSize,mWallSize,mWallSizeH,mWallThick,angle_);
        //console.log('slopeL1,2,3:'+[slope.L1, slope.L2, slope.L3]);
        mSetSlopeGrid(slope);

        return slope;
    }
    buildSlope(){
      
        this.mode = 2;
        this.buildType = 2;
        var slope = this.funcSlope();
        var j = mCheckUniqueBuild(slope,2);
        if(j){
            this.buildTemp = this.funcSlope();
            if(this.nowTouboBuild==1){
                this.doBuild();
            }//
        }//
        else{
            this.buildTemp = null;
        }
    }
    funcCone()
    { 
        var angle_ = this.angle; // + 3.1415/2;
        angle_ = Math.floor( (angle_+3.1415/4) / (3.1415/2)) * 3.1415/2;
 
        // player geo center
        var x_ = this.x;// 
        var y_ = this.y;// 
        x_ = mFloorGS(x_); 
        y_ = mFloorGS(y_); 

        var z_ = mFloorGSH(this.z + mWallSizeH*0.1);
        var dz_ =  this.z - mFloorGSH(this.z);
        //console.log('dz_:'+dz_+', angle2:'+this.angle2);
        if( (dz_>mFieldGridSizeH/2) && (this.angle2<-3.1415/6) && (this.angle2>-3.1415/4) ){
            z_ += mFieldGridSizeH;
        }

        x_ += mFieldGridSize * Math.cos(angle_);
        y_ += mFieldGridSize * Math.sin(angle_);

        if( (this.angle2>3.1415/4) ){
            x_ = mFloorGS(this.x);
            y_ = mFloorGS(this.y);
        }

        if( (this.angle2<=-3.1415/4) ){
            x_ = mFloorGS(this.x);
            y_ = mFloorGS(this.y);
            z_ = mFloorGSH(this.z + mFallSpeed*2) + mFieldGridSizeH;
        }


        let cone = mCreateCone(x_,y_,z_,mWallSize,mWallSize,mWallSizeH/2,mWallThick,angle_);
        //console.log('slopeL1,2,3:'+[slope.L1, slope.L2, slope.L3]);
        //mSetSlopeGrid(slope);

        var dw =  20;
        var Lgx = (mWallSize - dw)/2;
        var Lgy = (mWallSize - dw)/2;
        cone.gridSize[0] = (mWallSize - dw)/2;
        cone.gridSize[1] = (mWallSize - dw)/2;
        cone.gridSize[2] = 5;

        cone.gridL0 = cone.gridSize[0];
        cone.gridL1 = cone.gridSize[1];
        cone.gridL2 = cone.gridSize[2];
        
        let originVec = Array(4*3).fill(0);

        let gridP = [];
        let gridF = [];

        for(var i=0;i<2; i++){
            for(var j=0;j<2; j++){
                var x,y,z;
                x = j*cone.gridSize[0] + j*dw;
                y = mWallSize - (i+1)*cone.gridSize[1] - i*dw;
                z = 0; 

                x += cone.x;
                y += cone.y;
                z += cone.z;

                originVec[(i*2+j)*3+0] = x;
                originVec[(i*2+j)*3+1] = y;
                originVec[(i*2+j)*3+2] = z;

                gridP[(i*2+j)*4+0] = [x,y,z];
                gridP[(i*2+j)*4+1] = [x+Lgx,y,z];
                gridP[(i*2+j)*4+2] = [x+Lgx,y+Lgy,z];
                gridP[(i*2+j)*4+3] = [x,y+Lgy,z];

                var k = (i*2+j)*4;
                gridF[(i*2+j)*4+0] = [k+0,k+1,k+2];
                gridF[(i*2+j)*4+1] = [k+0,k+2,k+1];
                gridF[(i*2+j)*4+2] = [k+0,k+2,k+3];
                gridF[(i*2+j)*4+3] = [k+0,k+3,k+2];
            }//     
        } //
        cone.gridOrigin = originVec;
        cone.gridP = gridP;
        cone.gridF = gridF;

        return cone;
    }
    buildCone(){
      
        this.mode = 2;
        this.buildType = 3;
        var cone = this.funcCone();
        var j = mCheckUniqueBuild(cone,3);
        if(j){
            this.buildTemp = this.funcCone();
            if(this.nowTouboBuild==1){
                this.doBuild();
            }//
        }//
        else{
            this.buildTemp = null;
        }
    }
    doBuild()
    {
        if(this.buildType==0){
            var wall = this.funcWall();
            var j = mCheckUniqueBuild(wall,0);
            if(j){
                walls[wall.id] = wall;    
                let P = mCreateNodeOfRect(wall.x, wall.y, wall.z, wall.Lx, wall.Ly, wall.Lz);
                let F = mCreateFaceOfRect();
                mCreatePieceOfWall(wall, P, F);
                io.to(this.socketId).emit('didBuild',1.0);    
            }//
        }//
        else if(this.buildType==1){
            var floor = this.funcFloor();
            var j = mCheckUniqueBuild(floor,1);
            if(j){
                floors[floor.id] = floor;
                mCreatePieceOfFloor(floor);    //
                io.to(this.socketId).emit('didBuild',1.0);    
            }//
        }
        else if(this.buildType==2){
            var slope = this.funcSlope();
            var j = mCheckUniqueBuild(slope,2);
            if(j){
                slopes[slope.id] = slope;
                mCreatePieceOfSlope(slope);    //
                io.to(this.socketId).emit('didBuild',1.0);    
            }//
        }
        else if(this.buildType==3){
            var cone = this.funcCone();
            var j = mCheckUniqueBuild(cone,3);
            if(j){
                cones[cone.id] = cone;
                mCreatePieceOfCone(cone);   //
                io.to(this.socketId).emit('didBuild',1.0);    
            }//
        }
    
    }
    getRayOfEye(){

        var ray_ = new Object();
        ray_.x = this.x - this.cameraPosition[1]*Math.sin(this.angle);
        ray_.y = this.y + this.cameraPosition[1]*Math.cos(this.angle);
        ray_.z = this.z + this.cameraPosition[2];
        ray_.dx = Math.cos(-this.angle2) * Math.cos(this.angle);
        ray_.dy = Math.cos(-this.angle2) * Math.sin(this.angle);
        ray_.dz = Math.sin(-this.angle2);
        //console.log('getRayOfEye:'+[ray_.x, ray_.y, ray_.z, ray_.dx, ray_.dy, ray_.dz]);

        return ray_;

    }
    judgeEditMode(){

        console.log('judgeEditMode:');

        var judge_ = false;

        if(( this.mode == 3) && (this.buildType==0) ){

            let e_wall = walls[this.editBuildID];
            let s = e_wall.gridSelected;
            if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                (s[3]==1) && (s[4]==0) && (s[5]==0) &&
                (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 0);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                     (s[3]==0) && (s[4]==1) && (s[5]==0) &&
                     (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 1);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 2);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 3);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==0) && (s[5]==0) &&
                   (s[6]==1) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 4);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==1) && (s[5]==0) &&
                   (s[6]==0) && (s[7]==1) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 5);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 6);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==0) && (s[5]==0) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 7);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && (s[2]==1) &&
                   (s[3]==0) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 8);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==1) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 9);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==0) && (s[5]==0) &&
                   (s[6]==1) && (s[7]==1) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 10);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && (s[2]==1) &&
                   (s[3]==0) && (s[4]==0) && (s[5]==0) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 11);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==1) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==1) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 12);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==1) && (s[5]==0) &&
                   (s[6]==1) && (s[7]==1) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 13);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==0) && (s[5]==1) &&
                   (s[6]==1) && (s[7]==0) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 14);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && (s[2]==0) &&
                   (s[3]==0) && (s[4]==1) && (s[5]==0) &&
                   (s[6]==1) && (s[7]==1) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 15);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && (s[2]==1) &&
                   (s[3]==1) && (s[4]==1) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==0) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 16);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && (s[2]==1) &&
                   (s[3]==0) && (s[4]==1) && (s[5]==1) &&
                   (s[6]==0) && (s[7]==1) && (s[8]==1)  )
            {
                mCreateEditedWall(e_wall, 17);
                e_wall.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && (s[2]==0) &&
                   (s[3]==1) && (s[4]==1) && (s[5]==0) &&
                   (s[6]==1) && (s[7]==1) && (s[8]==0)  )
            {
                mCreateEditedWall(e_wall, 18);
                e_wall.isEditMode = 2;
            }
            else{
                e_wall.isEditMode = 0;
                e_wall.clearSubPiece();
                let P = mCreateNodeOfRect(e_wall.x, e_wall.y, e_wall.z, e_wall.Lx, e_wall.Ly, e_wall.Lz);
                let F = mCreateFaceOfRect();
                mCreatePieceOfWall(e_wall, P, F);
            }//

            if(e_wall.isEditChanged==0){
                e_wall.isEditMode = 0;
                e_wall.clearSubPiece();
                let P = mCreateNodeOfRect(e_wall.x, e_wall.y, e_wall.z, e_wall.Lx, e_wall.Ly, e_wall.Lz);
                let F = mCreateFaceOfRect();
                mCreatePieceOfWall(e_wall, P, F);
                this.resetSelectGrid();
            }
            
            this.mode = 2;
            this.editBuildID = -1;

            io.to(this.socketId).emit('didEdit',1.0);

            return judge_;
        }
        else if(( this.mode == 3) && (this.buildType==1) ){

            let e_floor = floors[this.editBuildID];
            let s = e_floor.gridSelected;

            if( (s[0]==1) && (s[1]==0) && 
                (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 0);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                     (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 1);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                     (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 2);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                     (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 3);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                     (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 4);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                     (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 5);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                     (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 6);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                     (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 7);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                     (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 8);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                     (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 9);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                     (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 10);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                     (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 11);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                     (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedFloor(e_floor, 12);
                e_floor.isEditMode = 2;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                     (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedFloor(e_floor, 13);
                e_floor.isEditMode = 2;
            }
            else{
                e_floor.isEditMode = 0;
                e_floor.clearSubPiece();
                mCreatePieceOfFloor(e_floor);
            }//

            if(e_floor.isEditChanged==0){
                e_floor.isEditMode = 0;
                e_floor.clearSubPiece();
                mCreatePieceOfFloor(e_floor);
                this.resetSelectGrid();
            }

            this.mode = 2;
            this.editBuildID = -1;

            io.to(this.socketId).emit('didEdit',1.0);

            return judge_;
        }
        else if(( this.mode == 3) && (this.buildType==2) ){

            let e_slope = slopes[this.editBuildID];
            let p = e_slope.pGridOrder;
            let d = e_slope.dGridOrder;

            if( (p[0]==0) && (p[1]==1) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 0);
                e_slope.isEditMode = 2;
            }
            else if( (p[0]==2) && (p[1]==3) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 1);
                e_slope.isEditMode = 2;
            }
            else if( (p[0]==1) && (p[1]==0) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 2);
                e_slope.isEditMode = 2;
            }            
            else if( (p[0]==3) && (p[1]==2) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 3);
                e_slope.isEditMode = 2;
            }      
            else if( (p[0]==2) && (p[1]==0) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 4);
                e_slope.isEditMode = 2;
            }            
            else if( (p[0]==3) && (p[1]==1) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 5);
                e_slope.isEditMode = 2;
            }  
            else if( (p[0]==0) && (p[1]==2) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 6);
                e_slope.isEditMode = 2;
            }            
            else if( (p[0]==1) && (p[1]==3) && (p.length==2)  )
            {
                mCreateEditedSlope(e_slope, 7);
                e_slope.isEditMode = 2;
            }  
            else if( (d[0]==0) && (d[1]==1) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 0);
            }   
            else if( (d[0]==0) && (d[1]==2) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 0);
            }   
            else if( (d[0]==0) && (d[1]==3) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 0);
            }   
            else if( (d[0]==1) && (d[1]==0) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415);
            }            
            else if( (d[0]==1) && (d[1]==2) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415);
            }            
            else if( (d[0]==1) && (d[1]==3) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415);
            } 
            else if( (d[0]==2) && (d[1]==3) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2);
            }            
            else if( (d[0]==2) && (d[1]==0) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2);
            }            
            else if( (d[0]==2) && (d[1]==1) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2);
            }            
            else if( (d[0]==3) && (d[1]==2) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2*3);
            }            
            else if( (d[0]==3) && (d[1]==0) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2*3);
            }            
            else if( (d[0]==3) && (d[1]==1) && (d.length==2)  )
            {
                mChangeSlopeDirection(e_slope, 3.1415/2*3);
            }            
            else{
                e_slope.isEditMode = 0;
                e_slope.clearSubPiece();
                mCreatePieceOfSlope(e_slope);
            }//

            if(e_slope.isEditChanged==0){
                e_slope.isEditMode = 0;
                e_slope.clearSubPiece();
                mCreatePieceOfSlope(e_slope);
                this.resetSelectGrid();
            }

            this.mode = 2;
            this.editBuildID = -1;

            io.to(this.socketId).emit('didEdit',1.0);

            return judge_;
        }
        else if(( this.mode == 3) && (this.buildType==3) ){

            let e_cone = cones[this.editBuildID];
            let s = e_cone.gridSelected;

            if( (s[0]==1) && (s[1]==0) && 
                (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 0);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                     (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 1);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                     (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 2);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                        (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 3);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                        (s[2]==0) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 4);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==0) && 
                        (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 5);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                        (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 6);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                        (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 7);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                        (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 8);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                        (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 9);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==0) && (s[1]==1) && 
                        (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 10);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==0) && 
                        (s[2]==1) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 11);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                        (s[2]==0) && (s[3]==1)   )
            {
                mCreateEditedCone(e_cone, 12);
                e_cone.isEditMode = 0;
            }
            else if( (s[0]==1) && (s[1]==1) && 
                        (s[2]==1) && (s[3]==0)   )
            {
                mCreateEditedCone(e_cone, 13);
                e_cone.isEditMode = 0;
            }
            else{
                e_cone.isEditMode = 0;
                for(var i=0;i<e_cone.subPieceNum;i++){
                    triangles[ e_cone.subPieceID[i] ].isEditMode = 0;
                }    
                e_cone.clearSubPiece();
                mCreatePieceOfCone(e_cone);
            }//
            
            if(e_cone.isEditChanged==0){
                e_cone.isEditMode = 0;
                e_cone.clearSubPiece();
                mCreatePieceOfCone(e_cone);
                this.resetSelectGrid();
            }

            this.mode = 2;
            this.editBuildID = -1;

            io.to(this.socketId).emit('didEdit',1.0);
            return judge_;
        }
        
        let ray_ = this.getRayOfEye();
        var judge_ = false;
        var min_e = mFieldGridSize*2;


        Object.values(triangles).forEach((triangle) => {

            console.log('triangle.parentType:'+triangle.parentType);

            var build;
            if(triangle.parentType==0){
                build = walls[triangle.parentID];
            }//
            else if(triangle.parentType==1){
                build = floors[triangle.parentID];
            }//
            else if(triangle.parentType==2){
                build = slopes[triangle.parentID];
            }//
            else if(triangle.parentType==3){
                build = cones[triangle.parentID];
            }//
            
            console.log('build.isEditMode:'+build.isEditMode);

            if( build.isEditMode != 1)
            {
                var res_ = mGetRayCross(ray_, triangle);
                //console.log('triangle['+triangle.id+']');
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);

                if( (res_.judge) && (res_.e<min_e) ){
                    judge_ = true;
                    min_e = res_.e;
                    this.mode = 3;
                    this.buildType = triangle.parentType;
                    this.editBuildID = build.id;
                }//
            }
        });    


        if(judge_){

            var id_ = this.editBuildID;
            //console.log('id_:'+id_);

            if(this.buildType==0){
                walls[id_].isEditMode = 1;
                console.log('subPieceNum:'+walls[id_].subPieceNum);
                for(var i=0;i<walls[id_].subPieceNum;i++){
                    console.log('  i:'+i);
                    triangles[ walls[id_].subPieceID[i] ].isEditMode = 1;
                }        
                walls[id_].isEditChanged = 0;
            }
            else if(this.buildType==1){
                floors[id_].isEditMode = 1;
                for(var i=0;i<floors[id_].subPieceNum;i++){
                    triangles[ floors[id_].subPieceID[i] ].isEditMode = 1;
                }        
                floors[id_].isEditChanged = 0;
            }
            else if(this.buildType==2){
                slopes[id_].isEditMode = 1;
                for(var i=0;i<slopes[id_].subPieceNum;i++){
                    triangles[ slopes[id_].subPieceID[i] ].isEditMode = 1;
                }        
                slopes[id_].isEditChanged = 0;
            }
            else if(this.buildType==3){
                cones[id_].isEditMode = 1;
                for(var i=0;i<cones[id_].subPieceNum;i++){
                    triangles[ cones[id_].subPieceID[i] ].isEditMode = 1;
                }        
                cones[id_].isEditChanged = 0;     
            }

            this.editBuildID = id_;
            console.log('Edit start, id_:'+id_);
            io.to(this.socketId).emit('startEdit',1.0);
        }//

        return judge_;
    }
    selectGridWall(){
        //console.log('selectGrid:'+this.editBuildID);

        let wall = walls[this.editBuildID];
        if(!wall){
            return;
        }
        //console.log('wall:'+[wall.id, wall.isSubPiece]);

        var p1_ = [0.0, 0.0, wall.gridL2];
        var p2_ = [0.0, wall.gridL1, 0.0];
        var d_ = [0,0,0];

        p1_ = mTranslatePhase(p1_, wall.M, d_);
        p2_ = mTranslatePhase(p2_, wall.M, d_);
        //console.log('p1_:'+p1_);

        let ray_ = this.getRayOfEye();
        
        for(var i=0;i<3; i++){
            for(var j=0;j<3; j++){
                               
                var grid_ = new Object();
                grid_.pvec = Array(9).fill(0);
                grid_.pvec[0] = wall.gridOrigin[(i*3+j)*3+0];
                grid_.pvec[1] = wall.gridOrigin[(i*3+j)*3+1];
                grid_.pvec[2] = wall.gridOrigin[(i*3+j)*3+2];
                grid_.pvec[3] = Math.abs(p1_[0]);
                grid_.pvec[4] = Math.abs(p1_[1]);
                grid_.pvec[5] = Math.abs(p1_[2]); 
                grid_.pvec[6] = Math.abs(p2_[0]);
                grid_.pvec[7] = Math.abs(p2_[1]);
                grid_.pvec[8] = Math.abs(p2_[2]);

                var res_ = mGetRayCross(ray_, grid_);
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);
                if(res_.judge){
                    wall.gridSelected[i*3+j] = 1;
                    wall.isEditChanged = 1;
                }//

            }     
        } 

    }
    selectGridFloor(){
        //console.log('selectGrid:'+this.editBuildID);

        let floor = floors[this.editBuildID];
        if(!floor){
            return;
        }
        let ray_ = this.getRayOfEye();

        var p1_ = [floor.gridL0, 0.0, 0.0];
        var p2_ = [0.0, floor.gridL1, 0.0];
        var d_ = [0,0,0];
        
        for(var i=0;i<2; i++){
            for(var j=0;j<2; j++){
               
                var grid_ = new Object();
                grid_.pvec = Array(9).fill(0);
                grid_.pvec[0] = floor.gridOrigin[(i*2+j)*3+0];
                grid_.pvec[1] = floor.gridOrigin[(i*2+j)*3+1];
                grid_.pvec[2] = floor.gridOrigin[(i*2+j)*3+2];
                grid_.pvec[3] = Math.abs(p1_[0]);
                grid_.pvec[4] = Math.abs(p1_[1]);
                grid_.pvec[5] = Math.abs(p1_[2]); 
                grid_.pvec[6] = Math.abs(p2_[0]);
                grid_.pvec[7] = Math.abs(p2_[1]);
                grid_.pvec[8] = Math.abs(p2_[2]);

                var res_ = mGetRayCross(ray_, grid_);
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);
                if(res_.judge){
                    floor.gridSelected[i*2+j] = 1;
                    floor.isEditChanged = 1;
                }//

            }     
        } 

    }
    selectGridSlope(){
        //console.log('selectGrid:'+this.editBuildID);

        let slope = slopes[this.editBuildID];
        if(!slope){
            return;
        }

        let ray_ = this.getRayOfEye();

        for(var i=0;i<2; i++){
            for(var j=0;j<2; j++){
               
                // Piece grid
                var grid_ = new Object();
                grid_.pvec = Array(9).fill(0);
                grid_.pvec[0] = slope.pGridOrigin[(i*2+j)*3+0];
                grid_.pvec[1] = slope.pGridOrigin[(i*2+j)*3+1];
                grid_.pvec[2] = slope.pGridOrigin[(i*2+j)*3+2];
                grid_.pvec[3] = slope.pGridSize[0];
                grid_.pvec[4] = 0;
                grid_.pvec[5] = 0; 
                grid_.pvec[6] = 0;
                grid_.pvec[7] = slope.pGridSize[1];
                grid_.pvec[8] = 0;
                grid_.type = 2;
                var res_ = mGetRayCross(ray_, grid_);
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);
                if(res_.judge){
                    
                    var n = slope.pGridOrder.length;
                    if(slope.pGridSelected[i*2+j]==0){
                        slope.pGridOrder[n] = i*2+j;
                        console.log('pGridOrder:'+slope.pGridOrder);    
                    }

                    slope.pGridSelected[i*2+j] = 1;
                    slope.isEditChanged = 1;
                }//

                // Direction grid
                grid_.pvec = Array(9).fill(0);
                grid_.pvec[0] = slope.dGridOrigin[(i*2+j)*3+0];
                grid_.pvec[1] = slope.dGridOrigin[(i*2+j)*3+1];
                grid_.pvec[2] = slope.dGridOrigin[(i*2+j)*3+2];
                grid_.pvec[3] = slope.dGridSize[(i*2+j)*3+0];
                grid_.pvec[4] = 0;
                grid_.pvec[5] = 0; 
                grid_.pvec[6] = 0;
                grid_.pvec[7] = slope.dGridSize[(i*2+j)*3+1];
                grid_.pvec[8] = 0;
                grid_.type = 2;

                res_ = mGetRayCross(ray_, grid_);
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);
                if(res_.judge){
                    var n = slope.dGridOrder.length;
                    if(slope.dGridSelected[i*2+j]==0){
                        slope.dGridOrder[n] = i*2+j;
                        console.log('dGridOrder:'+slope.dGridOrder);    
                    }
                    slope.dGridSelected[i*2+j] = 1;
                    slope.isEditChanged = 1;
                }//

            }     
        } 

    }
    selectGridCone(){
        //console.log('selectGrid:'+this.editBuildID);

        let cone = cones[this.editBuildID];
        if(!cone){
            return;
        }

        let ray_ = this.getRayOfEye();

        var p1_ = [cone.gridL0, 0.0, 0.0];
        var p2_ = [0.0, cone.gridL1, 0.0];
        var d_ = [0,0,0];
        
        for(var i=0;i<2; i++){
            for(var j=0;j<2; j++){
               
                //wall.gridOrigin[(i*3+j)*3+1] = wall.y + mWallSize - (i+1)*wall.gridSize[1] - i*dw;
                
                var grid_ = new Object();
                grid_.pvec = Array(9).fill(0);
                grid_.pvec[0] = cone.gridOrigin[(i*2+j)*3+0];
                grid_.pvec[1] = cone.gridOrigin[(i*2+j)*3+1];
                grid_.pvec[2] = cone.gridOrigin[(i*2+j)*3+2];
                grid_.pvec[3] = Math.abs(p1_[0]);
                grid_.pvec[4] = Math.abs(p1_[1]);
                grid_.pvec[5] = Math.abs(p1_[2]); 
                grid_.pvec[6] = Math.abs(p2_[0]);
                grid_.pvec[7] = Math.abs(p2_[1]);
                grid_.pvec[8] = Math.abs(p2_[2]);

                var res_ = mGetRayCross(ray_, grid_);
                //console.log('ray:res_:'+[res_.judge, res_.s, res_.t, res_.e]);
                if(res_.judge){
                    cone.gridSelected[i*2+j] = 1;
                    cone.isEditChanged = 1;
                }//

            }     
        } 

    }
    resetSelectGrid(){
        console.log('resetSelectGrid:'+this.editBuildID);
        if(this.editBuildID<0){
            return;
        }

        if(this.buildType==0){
            let wall = walls[this.editBuildID];
            for(var i=0;i<3*3; i++){         
                wall.gridSelected[i] = 0;
            }     
        }
        else if(this.buildType==1){
            let floor = floors[this.editBuildID];
            for(var i=0;i<2*2; i++){         
                floor.gridSelected[i] = 0;
            }     
        }
        else if(this.buildType==2){
            let slope = slopes[this.editBuildID];
            for(var i=0;i<4; i++){         
                slope.pGridSelected[i] = 0;
                slope.dGridSelected[i] = 0;
            }     
            slope.pGridOrder = [];
            slope.dGridOrder = [];
        }
        else if(this.buildType==3){
            let cone = cones[this.editBuildID];
            for(var i=0;i<2*2; i++){         
                cone.gridSelected[i] = 0;
            }     
        }

    }
    useBandage(){
        if((this.bandages>0) && this.health<this.maxHealth ){
            this.health += 10;
            this.health = Math.min(this.health,this.maxHealth);
            this.bandages -= 1;
            io.to(this.socketId).emit('usedBandage',1.0);
        }
        else{
            io.to(this.socketId).emit('cannotUseBandage',1.0);
        }
    }
    usePotion(){
        if((this.potions>0) && this.shield<this.maxShield ){
            this.shield += 50;
            this.shield = Math.min(this.shield,this.maxShield);
            this.potions -= 1;
            io.to(this.socketId).emit('usedPotion',1.0);
        }
        else{
            io.to(this.socketId).emit('cannotUsePotion',1.0);
        }
    }
    squat(){
        if(this.isSquat == 1){
            this.isSquat = 0;
            this.cameraPosition[2] = this.tall*0.8;
        } 
        else{
            this.isSquat = 1;
            this.cameraPosition[2] = this.tall*0.8*0.7;
        }
    }
    jump(){
        //this.z = 100;
        if(this.isBot==0){
            console.log('Jump, isGrounded:'+this.isGrounded);
        }
        
        if((this.isJump==0) && (this.isGrounded==1)){
            this.isGrounded = 0;
            this.isJump=1;
            this.jumpTime = this.timeCount;
            this.jumpStartZ = this.z;
        }
    }
    weaponChange(v_){
        this.weapon = v_;
        this.mode = 1;
    }
    emote(){
        if(this.isEmote == 1){
            this.isEmote = 0;
        } 
        else{
            this.isEmote = 1;
        }       
    }
    damage(v_){
        //this.health --;

        if(this.shield>0){
            this.shield -= v_;
            if(this.shield<0){
                v_ = -this.shield;
                this.shield = 0;
            }
            else{
                v_ = 0;
            }
        }

        this.health -= v_;
        this.health = Math.max(0,this.health);
        io.to(this.socketId).emit('recievedamage',1.0);
        if(this.health === 0){
            var id = this.lastDamegePlayer;
            let givePlayer = players[id];
            if(givePlayer){
                givePlayer.heal(50);
            }//
            this.remove();
        }
    }
    heal(v_){
       
        if(this.health<100){
            this.health += v_;
            if(this.health>100){
                v_ = this.health - 100;
                this.health = 100;
            }
            else{
                v_ = 0;
            }
        }

        this.shield += v_;
        this.shield = Math.min(100,this.shield);
    }
    stormDamage(v_){
        this.health -= v_;
        this.health = Math.max(0,this.health);
        //io.to(this.socketId).emit('recievedamage',1.0);
        if(this.health === 0){
            this.remove();
        }
    }
    remove(){
        io.sockets.emit('deadothers');
        delete players[this.id];
        io.to(this.socketId).emit('dead', (Object.keys(players).length+1) );
    }
    toJSON(){
        return Object.assign(super.toJSON(), {health: this.health, maxHealth: this.maxHealth, shield: this.shield, maxShield: this.maxShield, tall: this.tall,
            socketId: this.socketId, point: this.point, nickname: this.nickname, weapon: this.weapon, isShoot: this.isShoot,  isSquat: this.isSquat, isJump: this.isJump,  isGrounded: this.isGrounded,  cameraPosition: this.cameraPosition, 
            order: this.order, isSkinLoaded: this.isSkinLoaded, giveDamage: this.giveDamage, timeCount: this.timeCount, shootTime: this.shootTime, firingTime: this.firingTime, isBot: this.isBot, 
            bandages: this.bandages, potions: this.potions, isEmote: this.isEmote, mode: this.mode, buildTemp: this.buildTemp, buildType: this.buildType, inStorm: this.inStorm});
    }
}



function mAbsVec(vec_)
{
    var n = vec_.length;
    let a_ = Array(n).fill(0);

    for(var i=0;i<n;i++){
        a_[i] = Math.abs(vec_[i]);
    }//

    return a_;
}

function mTranslatePhase(vec_, M, d)
{
    //console.log('mTranslatePhase');

    var n = vec_.length;
    var out_ = Array(n).fill(0);

    for(var i=0;i<n/3;i++){
        var x = vec_[i*3+0];
        var y = vec_[i*3+1];
        var z = vec_[i*3+2];

        out_[i*3+0] = M[0][0]*x + M[0][1]*y + M[0][2]*z + d[0];
        out_[i*3+1] = M[1][0]*x + M[1][1]*y + M[1][2]*z + d[1];
        out_[i*3+2] = M[2][0]*x + M[2][1]*y + M[2][2]*z + d[2];
    }//

    return out_;
}//

function mSolveAb(A,b){

    var ans_ = new Object();
    ans_.detA = 0.0;
    ans_.x = 0.0;
    ans_.y = 0.0;
    ans_.z = 0.0;

    var a11 = A[0][0];
    var a12 = A[0][1];
    var a13 = A[0][2];
    var a21 = A[1][0];
    var a22 = A[1][1];
    var a23 = A[1][2];
    var a31 = A[2][0];
    var a32 = A[2][1];
    var a33 = A[2][2];

    var b1 = b[0];
    var b2 = b[1];
    var b3 = b[2];

    var detA_ = a11*a22*a33 + a12*a23*a31 + a13*a21*a32 - a13*a22*a31 - a12*a21*a33 - a11*a23*a32;
    if( Math.abs(detA_) < 0.000001){
        ans_.detA = 0;
        return ans_;
    }
    ans_.detA = detA_;
    //console.log(' detA:'+detA);

    ans_.x =(  (a22*a33-a23*a32)*b1 -(a12*a33-a13*a32)*b2 + (a12*a23-a13*a22)*b3 ) / detA_;
    ans_.y =( -(a21*a33-a23*a31)*b1 +(a11*a33-a13*a31)*b2 - (a11*a23-a13*a21)*b3 ) / detA_;
    ans_.z =(  (a21*a32-a22*a31)*b1 -(a11*a32-a12*a31)*b2 + (a11*a22-a12*a21)*b3 ) / detA_;

    return ans_;
}

function mGetRayCross(ray_, wall_){

    var res_ = new Object();
    res_.judge = false;
    res_.s = 0.0;
    res_.t = 0.0;
    res_.e = 0.0;

    var i = 0;
    if( (wall_.type==0) && (wall_.isSubPiece==1) && (wall_.Lx>wall_.Ly) ){
        i = 2;
    }//

    let A = [
        [wall_.pvec[i*9+3], wall_.pvec[i*9+6], -ray_.dx],
        [wall_.pvec[i*9+4], wall_.pvec[i*9+7], -ray_.dy],
        [wall_.pvec[i*9+5], wall_.pvec[i*9+8], -ray_.dz]
    ];

    let b = [ray_.x - wall_.pvec[i*9+0], ray_.y - wall_.pvec[i*9+1], ray_.z - wall_.pvec[i*9+2]];

    var ans_ = mSolveAb(A,b);
    res_.s = ans_.x;
    res_.t = ans_.y;
    res_.e = ans_.z;
    //console.log("ans_;"+[res_.s, res_.t, res_.e]);

    var z_max = mFieldGridSize;
    if(wall_.type==2){ //slope
        z_max = mFieldGridSize*1.8;
    }
    //console.log("z_max:"+[z_max, ans_.z, wall_.type]);

    if(wall_.type!=3){
        if( (ans_.detA!=0) && (ans_.x>=0) && (ans_.x<=1.0) && (ans_.y>=0) && (ans_.y<=1.0) && (ans_.z>=0) && (ans_.z<=z_max) ){
            res_.judge = true;
        }    
    }
    else if(wall_.type==3){
        if( (ans_.detA!=0) && (ans_.x>=0) && (ans_.y>=0) && (ans_.x+ans_.y<=1.0) && (ans_.z>=0) && (ans_.z<=z_max) ){
            res_.judge = true;
        }    
    }

    return res_;
   
}


function mCheckUniqueBuild(wall_temp, type_){

    var j = true;
    var err_ = 0.1;

    if(type_==0){
        Object.values(walls).forEach((wall) => {

            if( (wall_temp.x==wall.x) && (wall_temp.y==wall.y) && (wall_temp.z==wall.z) && 
                (wall_temp.Lx==wall.Lx) && (wall_temp.Lz==wall.Lz) && (wall_temp.Lz==wall.Lz) )
            {
                j = false;
                return j;
            }
        });    
    }
    else if(type_==1){
        Object.values(floors).forEach((floor) => {

            if( (wall_temp.x==floor.x) && (wall_temp.y==floor.y) && (wall_temp.z==floor.z) && 
                (wall_temp.Lx==floor.Lx) && (wall_temp.Lz==floor.Lz) && (wall_temp.Lz==floor.Lz) )
            {
                j = false;
                return j;
            }
        });    
    }
    else if( (type_==2) || (type_==3) ){
        Object.values(slopes).forEach((slope) => {

            var v =  Math.abs(wall_temp.gridCenter[0]-slope.gridCenter[0])
                    +Math.abs(wall_temp.gridCenter[1]-slope.gridCenter[1])
                    +Math.abs(wall_temp.gridCenter[2]-slope.gridCenter[2]);

            if( v < err_ )
            {
                j = false;
                return j;
            }
        });
        
        Object.values(cones).forEach((cone) => {

            var v =  Math.abs(wall_temp.gridCenter[0]-cone.gridCenter[0])
                    +Math.abs(wall_temp.gridCenter[1]-cone.gridCenter[1])
                    +Math.abs(wall_temp.gridCenter[2]-cone.gridCenter[2]);

            if( v < err_ )
            {
                j = false;
                return j;
            }

        });            

    }//
    
    return j;
}//


class Wall extends GameObject{
    constructor(obj={}){
        super(obj);
        this.health = this.maxHealth = 100;
        this.p0x = 0.0;
        this.p0y = 0.0;
        this.p0z = 0.0;
        this.p1x = 0.0;
        this.p1y = 0.0;
        this.p1z = 0.0;
        this.p2x = 0.0;
        this.p2y = 0.0;
        this.p2z = 0.0;
        this.pvec = Array(9*6).fill(0);
        this.nvec = Array(3*6).fill(0);
        this.isEditMode = 0;
        this.gridSize = Array(3).fill(0);
        this.gridOrigin = Array(3*9).fill(0);
        this.gridSelected = Array(9).fill(0);
        this.isEditChanged = 0;
        this.subPieceNum = 0;
        this.subPieceID = Array(2*6*5).fill(0);
        this.parentID = 0;
        this.isSubPiece = 0;
        this.gridL0 = 0;
        this.gridL1 = 0;
        this.gridL2 = 0;
        this.M = [[0,0,0],[0,0,0],[0,0,0]];
        this.offset = [0,0,0];
        this.type = 0;
        this.hitTime = 0;
        this.hitPosition = [0,0,0];
    }
    damage(v_){
        //this.health --;
        this.health -= v_;
        this.health = Math.max(0,this.health);
        //console.log('wall damage:'+this.id+', health:'+this.health);

        if(this.isSubPiece==0){
            var n = this.subPieceNum;
            for(var i=0;i<n; i++){
                var j = this.subPieceID[i];
                //walls[j].health = this.health;
                triangles[j].health = this.health;
            }//
        }

        if(this.health === 0){
            this.remove();
        }
    }
    clearSubPiece(){
        var n = this.subPieceNum;
        for(var i=0;i<n; i++){
            var j = this.subPieceID[i];
            delete triangles[j];
            console.log('triangle['+j+'] is deleted.');
            this.subPieceID[i] = 0;
        }//
        this.subPieceNum = 0;
    }
    remove(){

        if(this.isSubPiece==0){
            this.clearSubPiece();
        }//

        delete walls[this.id];
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, angle: this.angle, angle2: this.angle2, 
                isEditMode: this.isEditMode, gridSize: this.gridSize, gridOrigin: this.gridOrigin, gridSelected: this.gridSelected,
                health: this.health, hitTime: this.hitTime, hitPosition: this.hitPosition};
    }
}


function mCreateWall(x_,y_,z_,Lx_,Ly_,Lz_){

    const wall = new Wall({
       
    });
    wall.type = 0;
    wall.isEditMode = 0;

    wall.angle = 0; 
    wall.angle2 = 0;
    
    wall.x = x_;   
    wall.y = y_; 
    wall.z = z_;
  
    wall.Lx = Lx_;
    wall.Ly = Ly_;
    wall.Lz = Lz_;

    return wall;    
}


function mCreateFaceOfTriPrism()
{
    let F =[]; // 2 + 3x2
    //x-,x+, y or z, y or z, y or z,
    F[0] = [0,1,2];
    F[1] = [3,5,4];

    F[2] = [0,3,4];
    F[3] = [0,4,1];

    F[4] = [1,4,5];
    F[5] = [1,5,2];

    F[6] = [2,5,3];
    F[7] = [2,3,0];

    return F;
}


function mCreateNodeOfRect(x_,y_,z_,Lx_,Ly_,Lz_)
{
    let P =[];

    P[0] = [x_,y_,z_];
    P[1] = [x_+Lx_,y_,z_];
    P[2] = [x_+Lx_,y_,z_+Lz_];
    P[3] = [x_,y_,z_+Lz_];
   
    P[4] = [x_,y_+Ly_,z_];
    P[5] = [x_+Lx_,y_+Ly_,z_];
    P[6] = [x_+Lx_,y_+Ly_,z_+Lz_];
    P[7] = [x_,y_+Ly_,z_+Lz_];

    var j = 0;
    var d = Lx_/2;
    if(Lx_>Ly_){
        j = 1;
        d = Ly_/2;
    }

    for(var i=0;i<8;i++){
        P[i][j] -= d;
    }//

    return P;
}

function mCreateFaceOfRect()
{
    let F =[]; //6x2
    //y-,y+,x-,x+,z-,z+
    F[0] = [0,1,2];
    F[1] = [0,2,3];
    F[2] = [5,4,7];
    F[3] = [5,7,6];
    F[4] = [4,0,3];
    F[5] = [4,3,7];
    F[6] = [1,5,6];
    F[7] = [1,6,2];
    F[8] = [1,0,4];
    F[9] = [1,4,5];
    F[10] = [3,2,6];
    F[11] = [3,6,7];

    return F;
}


function mCreatePieceOfWall(wall, P, F)  //Parent wall, Node mat, Face mat
{ 
    var n = F.length;

    for(var i=0;i<n;i++)
    {
        let Ptri1 = [ P[F[i][0]],
                      P[F[i][1]],
                      P[F[i][2]]  ];
        let triangle1 = mCreateTriangle(Ptri1, 0);
        triangle1.isSubPiece = 1;
        triangle1.parentID = wall.id;
        triangle1.parentType = 0;
        //wall.subPieceID[i] = triangle1.id;
        wall.subPieceID[ wall.subPieceNum ] = triangle1.id;
        triangles[triangle1.id] = triangle1;
        wall.subPieceNum += 1;
    }// i

}//


function mCreateEditedWall(e_wall, type_)
{
    console.log("mCreatSubWall:");

    e_wall.clearSubPiece();

    var ox_ = e_wall.x;
    var oy_ = e_wall.y;
    var oz_ = e_wall.z;

    var Lw = [mWallThick, mWallSize, mWallSizeH];

    var n = 0;
    let p_ = [  [0,0,0], [0,0,0], [0,0,0], [0,0,0], [0,0,0]  ];
    let L_ = [  [0,0,0], [0,0,0], [0,0,0], [0,0,0], [0,0,0]  ];

    if(type_==0){ //
        // x x x
        //   x x  
        // x x x

        n=4
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/3*2, 0];
        p_[2] = [0, Lw[1]/3*2, Lw[2]/3*2];
        p_[3] = [0, Lw[1]/10*9, 0];
        L_[0] = [Lw[0],Lw[1]/3*2,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[3] = [Lw[0],Lw[1]/10,Lw[2]];
    }//
    else if(type_==1){ //
        // x x x
        // x   x  
        // x x x

        n=4
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/3, 0];
        p_[2] = [0, Lw[1]/3, Lw[2]/3*2];
        p_[3] = [0, Lw[1]/3*2, 0];
        L_[0] = [Lw[0],Lw[1]/3,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[3] = [Lw[0],Lw[1]/3,Lw[2]];
    }//
    else if(type_==2){ //
        // x x x
        // x x    
        // x x x

        n=4
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, 0];
        p_[2] = [0, 0, Lw[2]/3*2];
        p_[3] = [0, Lw[1]/3, 0];
        L_[0] = [Lw[0],Lw[1]/10,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[3] = [Lw[0],Lw[1]/3*2,Lw[2]];
    }//
    else if(type_==3){ //
        // x x x
        //   x    
        // x x x

        n=5;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, 0];
        p_[2] = [0, 0, Lw[2]/3*2];
        p_[3] = [0, Lw[1]/3, Lw[2]/3];
        p_[4] = [0, Lw[1]/10*9, 0];
        L_[0] = [Lw[0],Lw[1]/10,Lw[2]];
        L_[1] = [Lw[0],Lw[1],Lw[2]/3];
        L_[2] = [Lw[0],Lw[1],Lw[2]/3];
        L_[3] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[4] = [Lw[0],Lw[1]/10,Lw[2]];
    }//
    else if(type_==4){ //
        // x x x
        //   x x  
        //   x x

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/3*2, Lw[2]/3*2];
        p_[2] = [0, Lw[1]/10*9, 0];
        L_[0] = [Lw[0],Lw[1]/3*2,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/10,Lw[2]];
    }//
    else if(type_==5){ //
        // x x x
        // x   x  
        // x   x

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/3, Lw[2]/3*2];
        p_[2] = [0, Lw[1]/3*2, 0];
        L_[0] = [Lw[0],Lw[1]/3,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3,Lw[2]];
    }//
    else if(type_==6){ //
        // x x x
        // x x    
        // x x  

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, Lw[2]/3*2];
        p_[2] = [0, Lw[1]/3, 0];
        L_[0] = [Lw[0],Lw[1]/10,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3,Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3*2,Lw[2]];
    }//
    else if(type_==11){ //
        // 
        // x x x  
        // x x x

        n=1;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0], Lw[1], Lw[2]/3*2];
    }//
    else if(type_==12){ //right under square
        // x x x
        // x 
        // x 

        n=2
        e_wall.subPieceNum = n;
        p_[0] = [0, Lw[1]/3*2, 0];
        p_[1] = [0, 0, Lw[2]/3*2];
        L_[0] = [Lw[0],Lw[1]/3,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3*2,Lw[2]/3];
    }//
    else if(type_==13){ //right under square
        // x x x
        //     x 
        //     x 

        n=2
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/3, Lw[2]/3*2];
        L_[0] = [Lw[0],Lw[1]/3,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/3*2,Lw[2]/3];
    }//
    else if(type_==14){ //
        // x x x
        //   x    
        //   x  

        n=4;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, Lw[2]/3*2];
        p_[2] = [0, Lw[1]/3, 0];
        p_[3] = [0, Lw[1]/10*9, 0];
        L_[0] = [Lw[0],Lw[1]/10,Lw[2]];
        L_[1] = [Lw[0],Lw[1],Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/3,Lw[2]];
        L_[3] = [Lw[0],Lw[1]/10,Lw[2]];
    }//
    else if(type_==15){ //
        // x x x
        // x   x  
        //     

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, Lw[2]/3*2];
        p_[2] = [0, Lw[1]/10*9, 0];
        L_[0] = [Lw[0],Lw[1]/10,Lw[2]/3*2];
        L_[1] = [Lw[0],Lw[1],Lw[2]/3];
        L_[2] = [Lw[0],Lw[1]/10,Lw[2]/3*2];
    }//
    else if(type_==16){ //
        // 
        //   
        // x x x

        n=1;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0], Lw[1], Lw[2]/3];
    }//
    else if(type_==17){ //
        // x
        // x  
        // x 

        n=1;
        e_wall.subPieceNum = n;
        p_[0] = [0, Lw[1]/3*2, 0];
        L_[0] = [Lw[0], Lw[1]/3, Lw[2]];
    }//
    else if(type_==18){ //
        //     x
        //     x  
        //     x 

        n=1;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0], Lw[1]/3, Lw[2]];
    }//
    else if(type_==7){ //
        //     x
        //   x x  
        // x x x

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, Lw[2]];
        p_[2] = [0, Lw[1], 0];
        p_[3] = [Lw[0], 0, 0];
        p_[4] = [Lw[0], 0, Lw[2]];
        p_[5] = [Lw[0], Lw[1], 0];
    }//
    else if(type_==8){ //
        // x
        // x x  
        // x x x

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1], Lw[2]];
        p_[2] = [0, Lw[1], 0];
        p_[3] = [Lw[0], 0, 0];
        p_[4] = [Lw[0], Lw[1], Lw[2]];
        p_[5] = [Lw[0], Lw[1], 0];
    }//
    else if(type_==9){ //
        // x x x
        // x x  
        // x 

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, Lw[2]];
        p_[1] = [0, Lw[1], Lw[2]];
        p_[2] = [0, Lw[1], 0];
        p_[3] = [Lw[0], 0, Lw[2]];
        p_[4] = [Lw[0], Lw[1], Lw[2]];
        p_[5] = [Lw[0], Lw[1], 0];
    }//
    else if(type_==10){ //
        // x x x
        //   x x  
        //     x

        n=3;
        e_wall.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, 0, Lw[2]];
        p_[2] = [0, Lw[1], Lw[2]];
        p_[3] = [Lw[0], 0, 0];
        p_[4] = [Lw[0], 0, Lw[2]];
        p_[5] = [Lw[0], Lw[1], Lw[2]];
    }//


    e_wall.subPieceNum = 0;
    if( (type_>=7) && (type_<=10) ){ // Triangular Prism

        for(var k=0;k<6;k++){
            p_[k] = mTranslatePhase(p_[k], e_wall.M, e_wall.offset);
            p_[k] = [ ox_+p_[k][0], oy_+p_[k][1], oz_+p_[k][2] ];
        }//
       
        let F = mCreateFaceOfTriPrism();
        mCreatePieceOfWall(e_wall, p_, F);

        return;
    }
    else
    {
        for(var k=0;k<n;k++)
        {
            p_[k] = mTranslatePhase(p_[k], e_wall.M, e_wall.offset);
            L_[k] = mTranslatePhase(L_[k], e_wall.M, [0,0,0]);
    
            for(var i=0;i<3;i++){
                if(L_[k][i]<0){
                    L_[k][i] = Math.abs(L_[k][i]);
                    p_[k][i] -= L_[k][i];
                }
            }
                
            let P = mCreateNodeOfRect(ox_+p_[k][0], oy_+p_[k][1], oz_+p_[k][2], L_[k][0], L_[k][1], L_[k][2]);
            let F = mCreateFaceOfRect();
            mCreatePieceOfWall(e_wall, P, F);
    
        }//
        
        return;
    }//

}//


class Floor extends GameObject{
    constructor(obj={}){
        super(obj);
        this.health = this.maxHealth = 100;
        this.p0x = 0.0;
        this.p0y = 0.0;
        this.p0z = 0.0;
        this.p1x = 0.0;
        this.p1y = 0.0;
        this.p1z = 0.0;
        this.p2x = 0.0;
        this.p2y = 0.0;
        this.p2z = 0.0;
        this.pvec = Array(9*6).fill(0);
        this.nvec = Array(3*2).fill(0);
        this.isEditMode = 0;
        this.gridSize = Array(3).fill(0);
        this.gridOrigin = Array(3*4).fill(0);
        this.gridSelected = Array(4).fill(0);
        this.isEditChanged = 0;
        this.gridP = [];
        this.gridF = [];
        this.subPieceNum = 0;
        this.subPieceID = Array(4).fill(0);
        this.parentID = 0;
        this.isSubPiece = 0;
        this.gridL0 = 0;
        this.gridL1 = 0;
        this.gridL2 = 0;
        this.M = [[0,0,0],[0,0,0],[0,0,0]];
        this.offset = [0,0,0];
        this.type = 1;
        this.hitTime = 0;
        this.hitPosition = [0,0,0];
    }
    damage(v_){
        //this.health --;
        this.health -= v_;
        this.health = Math.max(0,this.health);
        //console.log('wall damage:'+this.id+', health:'+this.health);

        if(this.isSubPiece==0){
            var n = this.subPieceNum;
            for(var i=0;i<n; i++){
                var j = this.subPieceID[i];
                //floors[j].health = this.health;
                triangles[j].health = this.health;
            }//
        }

        if(this.health === 0){
            this.remove();
        }
    }
    clearSubPiece(){
        var n = this.subPieceNum;
        for(var i=0;i<n; i++){
            var j = this.subPieceID[i];
            delete triangles[j];
            console.log('Triangle['+j+'] is deleted.');
            this.subPieceID[i] = 0;
        }//
        this.subPieceNum = 0;
    }
    remove(){

        if(this.isSubPiece==0){
            this.clearSubPiece();
        }//

        delete floors[this.id];
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, angle: this.angle, angle2: this.angle2, 
                isEditMode: this.isEditMode, gridSize: this.gridSize, gridOrigin: this.gridOrigin, gridSelected: this.gridSelected, gridP: this.gridP, gridF: this.gridF,
                health: this.health, hitTime: this.hitTime, hitPosition: this.hitPosition };
    }
}


function mCreateFloor(x_,y_,z_,Lx_,Ly_,Lz_)
{
    //console.log("mCreatFloor:");

    const floor = new Floor({
       
    });
    floor.type = 1;
    floor.isEditMode = 0;

    floor.angle = 0; 
    floor.angle2 = 0;
    
    floor.x = x_;   
    floor.y = y_; 
    floor.z = z_;
  
    floor.Lx = Lx_;
    floor.Ly = Ly_;
    floor.Lz = Lz_;

    return floor;
}


function mCreatePieceOfFloor(floor){

    var x_ = floor.x;
    var y_ = floor.y;
    var z_ = floor.z;
    var Lx_ = floor.Lx;
    var Ly_ = floor.Ly;

    let Ptri1 = [ [x_,y_,z_],
                  [x_+Lx_,y_,z_],
                  [x_+Lx_,y_+Ly_,z_]  ];
    let triangle1 = mCreateTriangle(Ptri1, 1);
    triangle1.isSubPiece = 1;
    triangle1.parentID = floor.id;
    triangle1.parentType = 1;
    floor.subPieceID[0] = triangle1.id;
    triangles[triangle1.id] = triangle1;
    floor.subPieceNum += 1;

    let Ptri2 = [ [x_,y_,z_],
                  [x_+Lx_,y_+Ly_,z_],
                  [x_,y_+Ly_,z_]  ];
    let triangle2 = mCreateTriangle(Ptri2, 1);
    triangle2.isSubPiece = 1;
    triangle2.parentID = floor.id;
    triangle2.parentType = 1;
    floor.subPieceID[1] = triangle2.id;
    triangles[triangle2.id] = triangle2;
    floor.subPieceNum += 1;

}//


function mCreateEditedFloor(pFloor_, type_)
{
    console.log("mCreateEditedFloor:");

    pFloor_.clearSubPiece();

    var ox_ = pFloor_.x;
    var oy_ = pFloor_.y;
    var oz_ = pFloor_.z;

    var Lw = [mWallSize, mWallSize, mFloorVisualThick];//mFloorThick

    var n = 0;
    let p_ = [  [0,0,0], [0,0,0]  ];
    let L_ = [  [0,0,0], [0,0,0]  ];

    if(type_==0){ //
        //   x 
        // x x   
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [Lw[0]/2, Lw[1]/2, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//
    else if(type_==1){ //
        // x 
        // x x   
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//
    else if(type_==2){ //
        // x x
        //   x   
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [Lw[1]/2, 0, 0];
        p_[1] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/2,Lw[2]];
    }//    
    else if(type_==3){ //
        // x x
        // x   
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0],Lw[1]/2,Lw[2]];
    }//    
    else if(type_==4){ //
        //  
        // x x   
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
    }//
    else if(type_==5){ //
        // x x 
        //    
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
    }//
    else if(type_==6){ //
        //   x 
        //   x  
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [Lw[1]/2, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]];
    }//
    else if(type_==7){ //
        // x 
        // x  
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]];
    }//
    else if(type_==8){ //
        //   x
        // x     
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        p_[1] = [Lw[0]/2, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//    
    else if(type_==9){ //
        // x
        //   x     
        
        n=2
        pFloor_.subPieceNum = n;
        p_[0] = [Lw[0]/2, 0, 0];
        p_[1] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
        L_[1] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//        
    else if(type_==10){ //
        // x  
        //   
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//
    else if(type_==11){ //
        //   x  
        //   
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [Lw[1]/2, Lw[1]/2, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//
    else if(type_==12){ //
        //    
        // x  
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//
    else if(type_==13){ //
        //     
        //   x
        
        n=1;
        pFloor_.subPieceNum = n;
        p_[0] = [Lw[1]/2, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1]/2,Lw[2]];
    }//

    pFloor_.subPieceNum = 0;
    for(var k=0;k<n;k++)
    {        
        var x_ = ox_+p_[k][0];
        var y_ = oy_+p_[k][1];
        var z_ = oz_+p_[k][2];
        var Lx_ = L_[k][0];
        var Ly_ = L_[k][1];

        let Ptri1 = [ [x_,y_,z_],
                      [x_+Lx_,y_,z_],
                      [x_+Lx_,y_+Ly_,z_]  ];
        let triangle1 = mCreateTriangle(Ptri1, 1);
        triangle1.isSubPiece = 1;
        triangle1.parentID = pFloor_.id;
        triangle1.parentType = 1;
        pFloor_.subPieceID[k*2+0] = triangle1.id;
        triangles[triangle1.id] = triangle1;
        pFloor_.subPieceNum += 1;

        let Ptri2 = [ [x_,y_,z_],
                      [x_+Lx_,y_+Ly_,z_],
                      [x_,y_+Ly_,z_]  ];
        let triangle2 = mCreateTriangle(Ptri2, 1);
        triangle2.isSubPiece = 1;
        triangle2.parentID = pFloor_.id;
        triangle2.parentType = 1;
        pFloor_.subPieceID[k*2+1] = triangle2.id;
        triangles[triangle2.id] = triangle2;
        pFloor_.subPieceNum += 1;
    }//

}//


class Slope extends GameObject{
    constructor(obj={}){
        super(obj);
        this.health = this.maxHealth = 100;
        this.p0x = 0.0;
        this.p0y = 0.0;
        this.p0z = 0.0;
        this.p1x = 0.0;
        this.p1y = 0.0;
        this.p1z = 0.0;
        this.p2x = 0.0;
        this.p2y = 0.0;
        this.p2z = 0.0;
        this.pvec = Array(9*6).fill(0);
        this.nvec = Array(3*2).fill(0);
        this.size = Array(3).fill(0);
        this.M = [[0,0,0],[0,0,0],[0,0,0]];
        this.offset = [0,0,0];
        this.isEditMode = 0;
        this.pGridSize = Array(3).fill(0);
        this.pGridOrigin = Array(3*4).fill(0);
        this.pGridSelected = Array(4).fill(0);
        this.pGridOrder = [];
        this.dGridSize = Array(3*4).fill(0);
        this.dGridOrigin = Array(3*4).fill(0);
        this.dGridSelected = Array(4).fill(0);
        this.dGridOrder = [];
        this.isEditChanged = 0;
        this.subPieceNum = 0;
        this.subPieceID = Array(4).fill(0);
        this.parentID = 0;
        this.isSubPiece = 0;
        this.L1 = 0.0;
        this.L2 = 0.0;
        this.L3 = 0.0;
        this.type = 2;
        this.hitTime = 0;
        this.hitPosition = [0,0,0];
        this.gridCenter = [0,0,0];
    }
    damage(v_){
        //this.health --;
        this.health -= v_;
        this.health = Math.max(0,this.health);
        console.log(' damage:'+this.id+', v:'+v_+', health:'+this.health);

        if(this.isSubPiece==0){
            var n = this.subPieceNum;
            for(var i=0;i<n; i++){
                var j = this.subPieceID[i];
                //slopes[j].health = this.health;
                triangles[j].health = this.health;
            }//
        }

        if(this.health === 0){
            this.remove();
        }
    }
    clearSubPiece(){
        var n = this.subPieceNum;
        for(var i=0;i<n; i++){
            var j = this.subPieceID[i];
            // delete slopes[j];
            // console.log('slope['+j+'] is deleted.');
            delete triangles[j];
            console.log('triangle['+j+'] is deleted.');
            this.subPieceID[i] = 0;
        }//
        this.subPieceNum = 0;
    }
    remove(){

        if(this.isSubPiece==0){
            this.clearSubPiece();
        }//

        delete slopes[this.id];
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, L1:this.L1, L2:this.L2, L3:this.L3, angle: this.angle, angle2: this.angle2, 
                isEditMode: this.isEditMode, pGridSize: this.pGridSize, pGridOrigin: this.pGridOrigin, pGridSelected: this.pGridSelected,
                dGridSize: this.dGridSize, dGridOrigin: this.dGridOrigin, dGridSelected: this.dGridSelected, pvec:this.pvec,
                health: this.health, hitTime: this.hitTime, hitPosition: this.hitPosition };
    }
}


//function mCreatePieceSlope(x_,y_,z_,L1_,L2_,L3_,L4_,angle_)
function mCreateSlope(x_,y_,z_,L1_,L2_,L3_,L4_,angle_)
{
    //console.log("mCreatPieceSlope:");

    const slope = new Slope({
        
    });
    slope.type = 2;

    slope.L1 = L1_;
    slope.L2 = L2_;
    slope.L3 = L3_;

    slope.Lx = Math.sqrt(L1_*L1_ + L3_*L3_);
    slope.Ly = L2_;
    slope.Lz = L4_;

    slope.angle = angle_;
    var a_ = Math.acos(L1_/slope.Lx);
    slope.angle2 = -a_;
    //console.log('slope.angle:'+slope.angle);

    slope.x = x_; 
    slope.y = y_; 
    slope.z = z_; 
    
    let pvec_ = Array(9*2).fill(0.0);
    let nvec_ = Array(3*2).fill(0.0);
    let size_ = Array(3).fill(0.0);

    size_[0] = Math.sqrt(L1_*L1_ + L3_*L3_);
    size_[1] = L2_;
    size_[2] = L4_;


    if( (slope.angle==0) || (slope.angle==3.1415*2) )
    {
        let M = [
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ];
        slope.M = M;
        slope.offset = [0,0,0];
        
    }
    else if(slope.angle==3.1415)
    {    
        slope.angle2 = a_;
   
        let M = [
            [-1, 0, 0],
            [0, -1, 0],
            [0, 0, 1]
        ];
        slope.M = M;
        slope.offset = [mWallSize,mWallSize,0];

    }
    else if(slope.angle==3.1415/2)
    {    
        let M = [
            [0,-1, 0],
            [1, 0, 0],
            [0, 0, 1]
        ];
        slope.M = M;
        slope.offset = [mWallSize,0,0];
        
    }
    else if(slope.angle==3.1415/2*3){
        
        slope.angle2 = a_;
        
        let M = [
            [0, 1, 0],
            [-1, 0, 0],
            [0, 0, 1]
        ];
        slope.M = M;
        slope.offset = [0,mWallSize,0];
        
    }

    slope.nvec = mTranslatePhase(nvec_, slope.M, [0,0,0]);
    slope.size = mAbsVec( mTranslatePhase(size_, slope.M, [0,0,0]) );
    //console.log('size:'+slope.size);
    
    slope.pvec = mTranslatePhase(pvec_, slope.M, [0,0,0]);
    for(var i=0;i<2;i++){
        slope.pvec[9*i+0] = pvec_[9*i+0] + slope.offset[0];
        slope.pvec[9*i+1] = pvec_[9*i+1] + slope.offset[1];
        //slope.pvec[9*i+2] = slope.z + slope.offset[2];
    }

    slope.Lx = slope.size[0];
    slope.Ly = slope.size[1];
    slope.Lz = slope.size[2];

    slope.gridCenter = [slope.x+mWallSize/2,
                        slope.y+mWallSize/2,
                        slope.z-mWallSizeH/2];

    return slope;
}//


function mCreatePieceOfSlope(slope){

    console.log('mCreatePieceOfSlope, angle:'+slope.angle);

    var x_ = slope.x;
    var y_ = slope.y;
    var z_ = slope.z - mWallSizeH/2;
    var Lx_ = mWallSize;
    var Ly_ = mWallSize;
    var Lz_ = mWallSizeH;
    var angle_ = slope.angle;

    // counter clock wise
    let P0 = [x_,y_,z_]; //bottom of slope
    let P1 = [x_,y_,z_]; //top of slope
    let P2 = [x_,y_,z_]; //top of slope
    let P3 = [x_,y_,z_]; //bottom of slope

    if(angle_==0){
        P1 = [x_+Lx_,y_,z_+Lz_];
        P2 = [x_+Lx_,y_+Ly_,z_+Lz_];
        P3 = [x_,y_+Ly_,z_];
    }//
    else if(angle_==3.1415){
        P0 = [x_+Lx_,y_+Ly_,z_];
        P1 = [x_,y_+Ly_,z_+Lz_];
        P2 = [x_,y_,z_+Lz_];
        P3 = [x_+Lx_,y_,z_];
    }//
    else if(angle_==3.1415/2){
        P0 = [x_+Lx_,y_,z_];
        P1 = [x_+Lx_,y_+Ly_,z_+Lz_];
        P2 = [x_,y_+Ly_,z_+Lz_];
        P3 = [x_,y_,z_];
    }//
    else if(angle_==3.1415/2*3){
        P0 = [x_,y_+Ly_,z_];
        P1 = [x_,y_,z_+Lz_];
        P2 = [x_+Lx_,y_,z_+Lz_];
        P3 = [x_+Lx_,y_+Ly_,z_];
    }//

    let Ptri1 = [ P0, P1, P2 ];
    let triangle1 = mCreateTriangle(Ptri1, 1);
    triangle1.isSubPiece = 1;
    triangle1.parentID = slope.id;
    triangle1.parentType = 2;
    slope.subPieceID[0] = triangle1.id;
    triangles[triangle1.id] = triangle1;
    slope.subPieceNum += 1;

    let Ptri2 = [ P0, P2, P3  ];
    let triangle2 = mCreateTriangle(Ptri2, 1);
    triangle2.isSubPiece = 1;
    triangle2.parentID = slope.id;
    triangle2.parentType = 2;
    slope.subPieceID[1] = triangle2.id;
    triangles[triangle2.id] = triangle2;
    slope.subPieceNum += 1;

}//



function mSetSlopeGrid(slope)
{
    var t = mFloorVisualThick;// 10;
    var dd = 10;
    var dw =  (mWallSize - dd*2)/5;
    slope.pGridSize[0] = dw*2;
    slope.pGridSize[1] = dw*2;
    slope.pGridSize[2] = t;

    for(var i=0;i<2; i++){
        for(var j=0;j<2; j++){
            slope.pGridOrigin[(i*2+j)*3+0] = j*slope.pGridSize[0] + j*(dw+dd*2);
            slope.pGridOrigin[(i*2+j)*3+1] = mWallSize - (i+1)*slope.pGridSize[1] - i*(dw+dd*2);
            slope.pGridOrigin[(i*2+j)*3+2] = 0; 

            slope.pGridOrigin[(i*2+j)*3+0] += slope.x;
            slope.pGridOrigin[(i*2+j)*3+1] += slope.y;
            slope.pGridOrigin[(i*2+j)*3+2] += slope.z - mFieldGridSizeH/2;
        }//     
    } //

    slope.dGridSize[3*0+0] = dw*2;
    slope.dGridSize[3*0+1] = dw;
    slope.dGridSize[3*0+2] = t;
    slope.dGridSize[3*1+0] = dw*2;
    slope.dGridSize[3*1+1] = dw;
    slope.dGridSize[3*1+2] = t;

    slope.dGridSize[3*2+0] = dw;
    slope.dGridSize[3*2+1] = dw*2;
    slope.dGridSize[3*2+2] = t;
    slope.dGridSize[3*3+0] = dw;
    slope.dGridSize[3*3+1] = dw*2;
    slope.dGridSize[3*3+2] = t;

    slope.dGridOrigin[3*0+0] = 0;
    slope.dGridOrigin[3*0+1] = dw*2+dd;
    slope.dGridOrigin[3*0+2] = 0;
    slope.dGridOrigin[3*1+0] = mWallSize-dw*2;
    slope.dGridOrigin[3*1+1] = dw*2+dd;
    slope.dGridOrigin[3*1+2] = 0;

    slope.dGridOrigin[3*2+0] = dw*2 + dd;
    slope.dGridOrigin[3*2+1] = 0;
    slope.dGridOrigin[3*2+2] = 0;
    slope.dGridOrigin[3*3+0] = dw*2 + dd;
    slope.dGridOrigin[3*3+1] = dw*2 + dw + dd*2;
    slope.dGridOrigin[3*3+2] = 0;

    for(var i=0;i<4; i++){
        slope.dGridOrigin[i*3+0] += slope.x;
        slope.dGridOrigin[i*3+1] += slope.y;
        slope.dGridOrigin[i*3+2] += slope.z - mFieldGridSizeH/2;
    }

}//


function mChangeSlopeDirection(pSlope, angle_)
{
    var x_ = pSlope.x;
    var y_ = pSlope.y;
    var z_ = pSlope.z;

    let slope = mCreateSlope(x_,y_,z_,mWallSize,mWallSize,mWallSizeH,mWallThick,angle_);
    mCreatePieceOfSlope(slope);
    mSetSlopeGrid(slope);

    slopes[slope.id] = slope;

    pSlope.remove();

}


function mCreateEditedSlope(pSlope_, type_)
{
    console.log("mCreatEditedSlope:"+[pSlope_.id, type_]);

    pSlope_.clearSubPiece();

    var ox_ = pSlope_.x;
    var oy_ = pSlope_.y;
    var oz_ = pSlope_.z;

    var Lw = [mWallSize, mWallSize, mWallSizeH];

    var n = 0;
    let p_ = [  [0,0,0], [0,0,0]  ];
    let L_ = [  [0,0,0], [0,0,0]  ];
    let angle_ = 0;

    if(type_==0){ //
        // x > x 
        //     
        //

        n=1;
        pSlope_.subPieceNum = n;
        p_[0] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
    }//
    else if(type_==1){ //
        //  
        //     
        // x > x

        n=1;
        pSlope_.subPieceNum = n;
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
    }//
    else if(type_==2){ //
        // x < x 
        //     
        //

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [0, 0, 0];
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        p_[0] = [0, Lw[1]/2, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        angle_ = 3.1415; 
    }//
    else if(type_==3){ //
        //  
        //     
        // x < x

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [0, -Lw[1]/2, 0];
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        angle_ = 3.1415; 
    }//
    else if(type_==4){ //
        // x 
        // ^    
        // x 

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [-Lw[0]/2, 0, 0];
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]];
        angle_ = 3.1415/2; 
    }//
    else if(type_==5){ //
        //    x 
        //    ^    
        //    x 

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [0, 0, 0]; //Global
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]]; //Local
        p_[0] = [Lw[0]/2, 0, 0]; //Global
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]]; //Global
        angle_ = 3.1415/2; 
    }//
    else if(type_==6){ //
        // x 
        // v    
        // x 

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [0, 0, 0];
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]];
        p_[0] = [0, 0, 0];
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]];
        angle_ = 3.1415/2*3; 
    }//
    else if(type_==7){ //
        //    x 
        //    v    
        //    x 

        n=1;
        pSlope_.subPieceNum = n;
        // p_[0] = [Lw[0]/2, 0, 0]; //Global
        // L_[0] = [Lw[0],Lw[1]/2,Lw[2]]; //Local
        p_[0] = [Lw[0]/2, 0, 0]; //Global
        L_[0] = [Lw[0]/2,Lw[1],Lw[2]]; //Local
        angle_ = 3.1415/2*3; 
    }//

    pSlope_.subPieceNum = 0;
    for(var k=0;k<n;k++)
    {        
        var x_ = ox_+p_[k][0];
        var y_ = oy_+p_[k][1];
        var z_ = oz_+p_[k][2] - mWallSizeH/2;
        var Lx_ = L_[k][0];
        var Ly_ = L_[k][1];
        var Lz_ = L_[k][2];

        // counter clock wise
        let P0 = [x_,y_,z_]; //bottom of slope
        let P1 = [x_,y_,z_]; //top of slope
        let P2 = [x_,y_,z_]; //top of slope
        let P3 = [x_,y_,z_]; //bottom of slope

        if(angle_==0){
            P1 = [x_+Lx_,y_,z_+Lz_];
            P2 = [x_+Lx_,y_+Ly_,z_+Lz_];
            P3 = [x_,y_+Ly_,z_];
        }//
        else if(angle_==3.1415){
            P0 = [x_+Lx_,y_+Ly_,z_];
            P1 = [x_,y_+Ly_,z_+Lz_];
            P2 = [x_,y_,z_+Lz_];
            P3 = [x_+Lx_,y_,z_];
        }//
        else if(angle_==3.1415/2){
            P0 = [x_+Lx_,y_,z_];
            P1 = [x_+Lx_,y_+Ly_,z_+Lz_];
            P2 = [x_,y_+Ly_,z_+Lz_];
            P3 = [x_,y_,z_];
        }//
        else if(angle_==3.1415/2*3){
            P0 = [x_,y_+Ly_,z_];
            P1 = [x_,y_,z_+Lz_];
            P2 = [x_+Lx_,y_,z_+Lz_];
            P3 = [x_+Lx_,y_+Ly_,z_];
        }//


        let Ptri1 = [ P0, P1, P2 ];
        let triangle1 = mCreateTriangle(Ptri1, 1);
        triangle1.isSubPiece = 1;
        triangle1.parentID = pSlope_.id;
        triangle1.parentType = 2;
        pSlope_.subPieceID[k*2+0] = triangle1.id;
        triangles[triangle1.id] = triangle1;
        pSlope_.subPieceNum += 1;

        let Ptri2 = [ P0, P2, P3  ];
        let triangle2 = mCreateTriangle(Ptri2, 1);
        triangle2.isSubPiece = 1;
        triangle2.parentID = pSlope_.id;
        triangle2.parentType = 2;
        pSlope_.subPieceID[k*2+1] = triangle2.id;
        triangles[triangle2.id] = triangle2;
        pSlope_.subPieceNum += 1;

    }//

}//


class Cone extends GameObject{ //wall:0, floor:1, slope:2, cone:3
    constructor(obj={}){
        super(obj);
        this.health = this.maxHealth = 100;
        this.pvec = Array(9*6).fill(0);
        this.nvec = Array(3*2).fill(0);
        this.isEditMode = 0;
        this.gridSize = Array(3).fill(0);
        this.gridOrigin = Array(3*4).fill(0);
        this.gridSelected = Array(4).fill(0);
        this.isEditChanged = 0;
        this.gridP = [];
        this.gridF = [];
        this.subPieceNum = 0;
        this.subPieceID = Array(4).fill(0);
        this.parentID = 0;
        this.isSubPiece = 0;
        this.gridL0 = 0;
        this.gridL1 = 0;
        this.gridL2 = 0;
        this.M = [[0,0,0],[0,0,0],[0,0,0]];
        this.offset = [0,0,0];
        this.type = 1;
        this.hitTime = 0;
        this.hitPosition = [0,0,0];
        this.P = [[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0]];
        this.F = [[0,0,0],[0,0,0],[0,0,0],[0,0,0]];
        this.gridCenter = [0,0,0];
    }
    damage(v_){
        //this.health --;
        this.health -= v_;
        this.health = Math.max(0,this.health);
        //console.log('cone damage:'+this.id+', health:'+this.health);

        if(this.isSubPiece==0){
            var n = this.subPieceNum;
            for(var i=0;i<n; i++){
                var j = this.subPieceID[i];
                triangles[j].health = this.health;
            }//
        }

        if(this.health === 0){
            this.remove();
        }
    }
    clearSubPiece(){
        var n = this.subPieceNum;
        for(var i=0;i<n; i++){
            var j = this.subPieceID[i];
            delete triangles[j];
            console.log('triangle['+j+'] is deleted.');
            this.subPieceID[i] = 0;
        }//
        this.subPieceNum = 0;
    }
    remove(){

        if(this.isSubPiece==0){
            this.clearSubPiece();
        }//

        delete cones[this.id];
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, angle: this.angle, angle2: this.angle2, 
                isEditMode: this.isEditMode, gridSize: this.gridSize, gridOrigin: this.gridOrigin, gridSelected: this.gridSelected, gridP: this.gridP, gridF: this.gridF,
                health: this.health, hitTime: this.hitTime, hitPosition: this.hitPosition, P:this.P, F:this.F };
    }
}


class Triangle extends GameObject{
    constructor(obj={}){
        super(obj);
        this.health = this.maxHealth = 100;
        this.pvec = Array(9*2).fill(0);
        this.nvec = Array(3*2).fill(0);
        this.isEditMode = 0;
        this.gridSize = Array(3).fill(0);
        this.gridOrigin = Array(3*4).fill(0);
        this.gridSelected = Array(4).fill(0);
        this.subPieceNum = 0;
        this.subPieceID = Array(4).fill(0);
        this.parentID = 0;
        this.parentType = -1;
        this.isSubPiece = 0;
        this.gridL0 = 0;
        this.gridL1 = 0;
        this.gridL2 = 0;
        this.M = [[0,0,0],[0,0,0],[0,0,0]];
        this.offset = [0,0,0];
        this.type = 1;
        this.hitTime = 0;
        this.hitPosition = [0,0,0];
        this.P = [[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0]];
        this.F = [[0,0,0],[0,0,0],[0,0,0],[0,0,0]];
        this.flag_backFace = 0;
    }
    damage(v_){
        //this.health --;
        this.health -= v_;
        this.health = Math.max(0,this.health);
        //console.log('wall damage:'+this.id+', health:'+this.health);
        if(this.health === 0){
            this.remove();
        }
    }
    clearSubPiece(){
        var n = this.subPieceNum;
        for(var i=0;i<n; i++){
            var j = this.subPieceID[i];
            delete floors[j];
            console.log('floor['+j+'] is deleted.');
            this.subPieceID[i] = 0;
        }//
        this.subPieceNum = 0;
    }
    remove(){

        if(this.isSubPiece==0){
            this.clearSubPiece();
        }//

        delete floors[this.id];
    }
    toJSON(){
        return {id: this.id, x: this.x, y: this.y, z: this.z, Lx: this.Lx, Ly: this.Ly, Lz: this.Lz, angle: this.angle, angle2: this.angle2, 
                isEditMode: this.isEditMode, gridSize: this.gridSize, gridOrigin: this.gridOrigin, gridSelected: this.gridSelected,
                health: this.health, hitTime: this.hitTime, hitPosition: this.hitPosition, P:this.P, F:this.F };
    }
}


function mCreateCone(x_,y_,z_,Lx_,Ly_,Lz_,Lt_,angle_){
    
    const cone = new Cone({

    });
    
    cone.x = x_;
    cone.y = y_;
    cone.z = z_;
    cone.Lx = Lx_; //mFieldGridSize;
    cone.Ly = Ly_; //mFieldGridSize;
    cone.Lz = Lz_; //mFieldGridSizeH/2;
    cone.gridCenter = [x_+Lx_/2, y_+Ly_/2, z_];

    let P =  [
        [x_, y_, z_], //p0
        [x_+Lx_, y_, z_], //p1
        [x_+Lx_, y_+Ly_, z_], //p2
        [x_, y_+Ly_, z_], //p3
        [x_+Lx_/2, y_+Ly_/2, z_+Lz_] //p4
    ]; 
    let F =  [
        [0, 1, 4], //f0
        [1, 2, 4], //f1
        [2, 3, 4], //f2
        [3, 0, 4] //f3
    ]; 
    cone.P = P;
    cone.F = F;

    cone.subPieceNum = 0;

    //cones[cone.id] = cone;
     return cone;
}

function mCreatePieceOfCone(cone){

    let P = cone.P;
    let F = cone.F;

    cone.subPieceNum = 0;
    for(let i=0; i<4; i++){

        var p0 = F[i][0];
        var p1 = F[i][1];
        var p2 = F[i][2];
        let Ptri = [P[p0],P[p1],P[p2]];
        let triangle = mCreateTriangle(Ptri, 1);

        triangle.isSubPiece = 1;
        triangle.parentID = cone.id;
        triangle.parentType = 3;
        triangles[triangle.id] = triangle;

        cone.subPieceID[i] = triangle.id;
        cone.subPieceNum += 1;
    }//

}//

function mCreateTriangle(P, flag_backFace){
    //P: 3x3
    const triangle = new Triangle({
    });
    triangle.type = 3;
    triangle.flag_backFace = flag_backFace;

    triangle.P = P;
    triangle.F[0] = [0,1,2];
    triangle.F[1] = [0,2,1];

    var p0 = triangle.F[0][0];
    var p1 = triangle.F[0][1];
    var p2 = triangle.F[0][2];
    triangle.pvec[9*0+0] = P[p0][0];
    triangle.pvec[9*0+1] = P[p0][1];
    triangle.pvec[9*0+2] = P[p0][2];
    triangle.pvec[9*0+3] = P[p1][0] - P[p0][0];
    triangle.pvec[9*0+4] = P[p1][1] - P[p0][1];
    triangle.pvec[9*0+5] = P[p1][2] - P[p0][2];
    triangle.pvec[9*0+6] = P[p2][0] - P[p0][0];
    triangle.pvec[9*0+7] = P[p2][1] - P[p0][1];
    triangle.pvec[9*0+8] = P[p2][2] - P[p0][2];

    var a = new Object();
    a.x = triangle.pvec[9*0+3];
    a.y = triangle.pvec[9*0+4];
    a.z = triangle.pvec[9*0+5];
    var b = new Object();
    b.x = triangle.pvec[9*0+6];
    b.y = triangle.pvec[9*0+7];
    b.z = triangle.pvec[9*0+8];
    let c = mOuterProduct(a,b);
    if( (Math.abs(c.x)+Math.abs(c.y) > 0) && (c.z>0) ){
        c.x = 0;
        c.y = 0;
    }
    c = mNormalize(c);
    console.log('c:'+[c.x, c.y, c.z]);

    triangle.nvec[3*0+0] = c.x;
    triangle.nvec[3*0+1] = c.y;
    triangle.nvec[3*0+2] = c.z;


    if(flag_backFace==1)
    {
        triangle.pvec[9*1+0] = P[p0][0];
        triangle.pvec[9*1+1] = P[p0][1];
        triangle.pvec[9*1+2] = P[p0][2];
        triangle.pvec[9*1+3] = P[p2][0] - P[p0][0];
        triangle.pvec[9*1+4] = P[p2][1] - P[p0][1];
        triangle.pvec[9*1+5] = P[p2][2] - P[p0][2];
        triangle.pvec[9*1+6] = P[p1][0] - P[p0][0];
        triangle.pvec[9*1+7] = P[p1][1] - P[p0][1];
        triangle.pvec[9*1+8] = P[p1][2] - P[p0][2];

        //var a = new Object();
        a.x = triangle.pvec[9*1+3];
        a.y = triangle.pvec[9*1+4];
        a.z = triangle.pvec[9*1+5];
        //var b = new Object();
        b.x = triangle.pvec[9*1+6];
        b.y = triangle.pvec[9*1+7];
        b.z = triangle.pvec[9*1+8];
        let c = mOuterProduct(a,b);
        if( Math.abs(c.x) + Math.abs(c.y) > 0 ){
            c.z = 0;
        }
        c = mNormalize(c);
        console.log('c:'+[c.x, c.y, c.z]);

        triangle.nvec[3*1+0] = c.x;
        triangle.nvec[3*1+1] = c.y;
        triangle.nvec[3*1+2] = c.z;
    }//

    return triangle;
}


function mCreateEditedCone(pCone, type_)
{
    console.log("mCreatEditedCone:");

    pCone.clearSubPiece();

    var ox_ = pCone.x;
    var oy_ = pCone.y;
    var oz_ = pCone.z;

    var Lw = [mWallSize, mWallSize, mWallSizeH];

    var n = 0;
    // let p_ = [  [0,0,0], [0,0,0]  ];
    // let L_ = [  [0,0,0], [0,0,0]  ];
    let P = [  [0,0,0], [0,0,0], [0,0,0],  [0,0,0], [0,0,0], [0,0,0]  ];

    if(type_==0){ //
        // s x 
        // x x   
        
        n=2
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, 0];
        P[3*0+2] = [0, Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], 0];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//
    else if(type_==1){ //
        // x s
        // x x   
        
        n=2
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, 0];
        P[3*0+2] = [Lw[0], Lw[1], Lw[2]];

        P[3*1+0] = [0, Lw[1], 0];
        P[3*1+1] = [0, 0, 0];
        P[3*1+2] = [Lw[0], Lw[1], Lw[2]];
    }//
    else if(type_==2){ //
        // x x
        // s x   
        
        n=2
        pCone.subPieceNum = n;
        P[3*0+0] = [Lw[0], 0, 0];
        P[3*0+1] = [Lw[0], Lw[1], 0];
        P[3*0+2] = [0, 0, Lw[2]];

        P[3*1+0] = [Lw[0], Lw[1], 0];
        P[3*1+1] = [0, Lw[1], 0];
        P[3*1+2] = [0, 0, Lw[2]];
    }//    
    else if(type_==3){ //
        // x x
        // x s 
        
        n=2
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [0, Lw[1], 0];

        P[3*1+0] = [Lw[0], Lw[1], 0];
        P[3*1+1] = [0, Lw[1], 0];
        P[3*1+2] = [Lw[0], 0, Lw[2]];
    }//    
    else if(type_==4){ //
        // s s
        // x x   
        
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, 0];
        P[3*0+2] = [0, Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//
    else if(type_==5){ //
        // x x 
        // s s   
        
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, Lw[2]];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [0, Lw[1], 0];

        P[3*1+0] = [Lw[0], 0, Lw[2]];
        P[3*1+1] = [Lw[0], Lw[1], 0];
        P[3*1+2] = [0, Lw[1], 0];
    }//
    else if(type_==6){ //
        // s x 
        // s x  
        
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, Lw[2]];
        P[3*0+1] = [Lw[0], 0, 0];
        P[3*0+2] = [0, Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], 0];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//
    else if(type_==7){ //
        // x s
        // x s  
        
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [0, Lw[1], 0];

        P[3*1+0] = [Lw[0], 0, Lw[2]];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, Lw[1], 0];
    }//
    else if(type_==8){ //
        // s x
        // x s    
        n=2
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [0, Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], Lw[1], 0];
        P[3*1+1] = [0, Lw[1], Lw[2]];
        P[3*1+2] = [Lw[0], 0, Lw[2]];
    }//    
    else if(type_==9){ //
        // x s
        // s x            
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, Lw[1], 0];
        P[3*0+1] = [0, 0, Lw[2]];
        P[3*0+2] = [Lw[0], Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, 0, Lw[2]];
    }//        
    else if(type_==10){ //
        // x s
        // s s        
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, Lw[2]];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [0, Lw[1], 0];

        P[3*1+0] = [Lw[0], 0, Lw[2]];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, Lw[1], 0];
    }//
    else if(type_==11){ //
        // s x  
        // s s 
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, Lw[2]];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [Lw[0], Lw[1], 0];

        P[3*1+0] = [0, 0, Lw[2]];
        P[3*1+1] = [Lw[0], Lw[1], 0];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//
    else if(type_==12){ //
        // s s  
        // x s 
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, 0];
        P[3*0+1] = [Lw[0], 0, Lw[2]];
        P[3*0+2] = [Lw[0], Lw[1], Lw[2]];

        P[3*1+0] = [0, 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//
    else if(type_==13){ //
        // s s   
        // s x
        n=2;
        pCone.subPieceNum = n;
        P[3*0+0] = [0, 0, Lw[2]];
        P[3*0+1] = [Lw[0], 0, 0];
        P[3*0+2] = [0, Lw[1], Lw[2]];

        P[3*1+0] = [Lw[0], 0, 0];
        P[3*1+1] = [Lw[0], Lw[1], Lw[2]];
        P[3*1+2] = [0, Lw[1], Lw[2]];
    }//


    for(var k=0;k<n;k++)
    {
        let Ptri = [ P[k*3+0], P[k*3+1], P[k*3+2] ];
        for(var i=0;i<3;i++){
            Ptri[i][0] += ox_;
            Ptri[i][1] += oy_;
            Ptri[i][2] += oz_;
        }//

        let triangle_1 = mCreateTriangle( Ptri, 1 );
        triangle_1.isSubPiece = 1;
        triangle_1.parentID = pCone.id;
        triangle_1.parentType = 3;
        pCone.subPieceID[k] = triangle_1.id;
        triangles[triangle_1.id] = triangle_1;    
    }//

}//


class Bandage extends GameObject{
    constructor(obj={}){
        super(obj);
        this.width = mBandageSize;
        this.height = mBandageSize;
        this.tall =  mBandageSize*2;
    }
    remove(){
        delete bandages[this.id];
    }
}

class Potion extends GameObject{
    constructor(obj={}){
        super(obj);
        this.width = mBandageSize;
        this.height = mBandageSize;
        this.tall =  mBandageSize*2;
    }
    remove(){
        delete potions[this.id];
    }
}

class Storm extends GameObject{
    constructor(obj={}){
        super(obj);
    }

}

var mStorm = 
    new Storm({
        x: 0,
        y: 0,
        z: 0,
        width: FIELD_WIDTH,
        height: FIELD_WIDTH,
        tall: FIELD_WIDTH,
        angle: 0,
        angle2: 0,
        isSquat: 0,
    });


function mBulletWallCollision(wall_, i, bullet_)
{
    var res = new Object();
    res.judge = false;
    res.e = -1.0;
    res.px = 0.0;
    res.py = 0.0;
    res.pz = 0.0;

    let A = [
        [wall_.pvec[i*9+3], wall_.pvec[i*9+6], -bullet_.dx],
        [wall_.pvec[i*9+4], wall_.pvec[i*9+7], -bullet_.dy],
        [wall_.pvec[i*9+5], wall_.pvec[i*9+8], -bullet_.dz]
    ];
    let b = [bullet_.x - wall_.pvec[i*9+0], 
            bullet_.y - wall_.pvec[i*9+1], 
            bullet_.z - wall_.pvec[i*9+2]];
    var ans_ = mSolveAb(A,b);
    //console.log('ans_:'+[ans_.x, ans_.y, ans_.z]);
    var s = ans_.x;
    var t = ans_.y;
    var e = ans_.z;
    
    if( (s>=0.0) &&  (t>=0.0) && (s+t<=1.0) && (e>=0.0)  ){
        res.judge = true;
        res.e = e;
        res.px = wall_.pvec[i*9+0] + wall_.pvec[i*9+3]*s + wall_.pvec[i*9+6]*t;
        res.py = wall_.pvec[i*9+1] + wall_.pvec[i*9+4]*s + wall_.pvec[i*9+7]*t;
        res.pz = wall_.pvec[i*9+2] + wall_.pvec[i*9+5]*s + wall_.pvec[i*9+8]*t;
    }    
    //console.log(' res:'+res);

    return res;
}


function mBulletPlayerCollision(player_, bullet_)
{
    //var plane1;
    const plane1 = new Wall({
        x: 0,
        y: 0,
        z: 0,
    });
    var t_ = player_.getHight(); 

    // y-z plane
    plane1.pvec[0] = player_.x;
    plane1.pvec[1] = player_.y - player_.width/2;
    plane1.pvec[2] = player_.z;
    plane1.pvec[3] = 0.0;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = t_;
    plane1.pvec[6] = 0.0;
    plane1.pvec[7] = player_.width;
    plane1.pvec[8] = 0.0;

    var res1 =mBulletWallCollision(plane1, 0, bullet_);
    if(res1.judge){
        return res1;
    }

    plane1.pvec[0] = player_.x;
    plane1.pvec[1] = player_.y + player_.width/2;
    plane1.pvec[2] = player_.z;
    plane1.pvec[3] = 0.0;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = t_;
    plane1.pvec[6] = 0.0;
    plane1.pvec[7] = -player_.width;
    plane1.pvec[8] = t_;
    
    var res1 =mBulletWallCollision(plane1, 0, bullet_);
    if(res1.judge){
        return res1;
    }

    // x-z plane
    plane1.pvec[0] = player_.x - player_.width/2;
    plane1.pvec[1] = player_.y;
    plane1.pvec[2] = player_.z;
    plane1.pvec[3] = 0.0;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = t_;
    plane1.pvec[6] = player_.width;
    plane1.pvec[7] = 0.0;
    plane1.pvec[8] = 0.0;

    res1 =mBulletWallCollision(plane1, 0, bullet_);
    if(res1.judge){
        return res1;
    }

    plane1.pvec[0] = player_.x + player_.width/2;
    plane1.pvec[1] = player_.y;
    plane1.pvec[2] = player_.z;
    plane1.pvec[3] = 0.0;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = t_;
    plane1.pvec[6] = -player_.width;
    plane1.pvec[7] = 0.0;
    plane1.pvec[8] = t_;

    res1 =mBulletWallCollision(plane1, 0, bullet_);
    if(res1.judge){
        return res1;
    }

    // x-y plane
    plane1.pvec[0] = player_.x - player_.width/2;
    plane1.pvec[1] = player_.y - player_.width/2;
    plane1.pvec[2] = player_.z + t_;
    plane1.pvec[3] = player_.width;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = 0.0;
    plane1.pvec[6] = 0.0;
    plane1.pvec[7] = player_.width;
    plane1.pvec[8] = 0.0;

    res1 =mBulletWallCollision(plane1, 0, bullet_);
    if(res1.judge){
        return res1;
    }

    plane1.pvec[0] = player_.x + player_.width/2;
    plane1.pvec[1] = player_.y + player_.width/2;
    plane1.pvec[2] = player_.z + t_;
    plane1.pvec[3] = -player_.width;
    plane1.pvec[4] = 0.0;
    plane1.pvec[5] = 0.0;
    plane1.pvec[6] = 0.0;
    plane1.pvec[7] = -player_.width;
    plane1.pvec[8] = 0.0;

    res1 =mBulletWallCollision(plane1, 0, bullet_);

    return res1;
}


function mNormalize(d_)
{
    var dmag_ = Math.sqrt(d_.x*d_.x + d_.y*d_.y + d_.z*d_.z);
    if(dmag_>0){
        d_.x = d_.x / dmag_;
        d_.y = d_.y / dmag_;
        d_.z = d_.z / dmag_;
    }
    else{

    }

    return d_;
}

function mOuterProduct(a_, b_)
{
    var c_ = new Object();
    c_.x = a_.y*b_.z - a_.z*b_.y;
    c_.y = a_.z*b_.x - a_.x*b_.z;
    c_.z = a_.x*b_.y - a_.y*b_.x;

    return c_;
}

function mMagVec(vec_)
{
    var m = 0;
    var n = vec_.length;
    for(var i=0;i<n;i++){
        m += vec_[i]*vec_[i];
    }//

    return Math.sqrt(m);
}


function mPlaneEdgePlayerCollisionDepth(plane_, i, player_)
{
    var res_ = new Object();
    res_.judge = false;
    res_.s = 0.0;
    res_.t = 0.0;
    res_.e = 0.0;
    res_.dx = 0.0;
    res_.dy = 0.0;

    var n_ = new Object();
    n_.x = 0.0;
    n_.y = 0.0;
    n_.z = 1.0;
   
    let A = [
        [plane_.pvec[i*9+3], plane_.pvec[i*9+6], -n_.x],
        [plane_.pvec[i*9+4], plane_.pvec[i*9+7], -n_.y],
        [plane_.pvec[i*9+5], plane_.pvec[i*9+8], -n_.z]
    ];
    let b = [player_.x - plane_.pvec[i*9+0], 
             player_.y - plane_.pvec[i*9+1], 
             player_.z - plane_.pvec[i*9+2]];
    var ans_ = mSolveAb(A,b);

    var z_min = 10;
    var z_max = player_.getHight();

    if( (ans_.x>=0.0) && (ans_.x+ans_.y<=1.0) &&
        (ans_.y>=0.0) && 
        (ans_.z>z_min) && (ans_.z<z_max) )
    {
        res_.judge = true;
        console.log('triangle edge depth:'+ ans_.z);
    }

    return res_;
}


function mPlanePointCollisionDepth(wall_, array_i, player_)
{
    //console.log('mPlanePointCollisionDepth:'+ array_i.length);

    var res_ = new Object();
    res_.judge = false;
    res_.s = 0.0;
    res_.t = 0.0;
    res_.e = 0.0;
    res_.isGrounded = player_.isGrounded;
    res_.i = 0;

    //var i = array_i[0]; 
    for(var j=0;j<array_i.length;j++)
    {
        var i = array_i[j]; 
        //console.log('  index:'+ i);

        var a_ = new Object();
        a_.x = wall_.pvec[i*9+3]; //plane_.p1x;
        a_.y = wall_.pvec[i*9+4]; //plane_.p1y;
        a_.z = wall_.pvec[i*9+5]; //plane_.p1z;
        var b_ = new Object();
        b_.x = wall_.pvec[i*9+6]; //plane_.p2x;
        b_.y = wall_.pvec[i*9+7]; //plane_.p2y;
        b_.z = wall_.pvec[i*9+8]; //plane_.p2z;

        var n_ = mOuterProduct(a_,b_);
        n_ = mNormalize(n_);

        res_.nx = wall_.nvec[i*3+0];
        res_.ny = wall_.nvec[i*3+1];
        res_.nz = wall_.nvec[i*3+2];
        //console.log('res_.n:'+[res_.nx, res_.ny, res_.nz]);

        var dz = 0;
        if( (wall_.type!=0) && (i==0) ){
            //dz = player_.tall;
            dz = player_.getHight();
        }
        // if(wall_.type==0){
        //     dz = player_.tall*0.1;
        // }
        if( (wall_.type==3) && (i==0)){ //foot to ground
            dz = player_.tall*0.1;
        }
        if( (wall_.type==3) && (i==1)){ //head to ceil
            dz = player_.getHight();
        }
        if( res_.nz==-1.0){
            dz = player_.getHight();
        }

        let A = [
            [wall_.pvec[i*9+3], wall_.pvec[i*9+6], -n_.x],
            [wall_.pvec[i*9+4], wall_.pvec[i*9+7], -n_.y],
            [wall_.pvec[i*9+5], wall_.pvec[i*9+8], -n_.z]
        ];
        let b = [player_.x - wall_.pvec[i*9+0], 
                 player_.y - wall_.pvec[i*9+1], 
                 player_.z + dz - wall_.pvec[i*9+2]];
        var ans_ = mSolveAb(A,b);
        //console.log('ans_:'+[ans_.x, ans_.y, ans_.z]);

        res_.s = ans_.x;
        res_.t = ans_.y;
        res_.e = ans_.z;

        var depth_ = mWallThick;
        if(wall_.type==0){
            depth_ = mWallThick/3;
        }

        if(wall_.flag_backFace==0){
            depth_ = mWallThick/3;
        }

            
        if( (ans_.x>=0.0) && (ans_.x+ans_.y<=1.0) &&
            (ans_.y>=0.0) && 
            (ans_.z>0) && (ans_.z<depth_) )
        {
            //console.log('ans_:'+[ans_.x, ans_.y, ans_.z]);
            //console.log('    :'+[wall_.pvec[0], wall_.pvec[1], wall_.pvec[2], wall_.id, wall_.parentID]);
            res_.judge = true;
            res_.i = i;
            if(res_.nz==1.0){
                res_.isGrounded = 1;
            }
            return res_;
        }

    }// for j

    return res_;
}



class Bullet extends GameObject{
    constructor(obj){
        super(obj);
        this.width = mBulletLength;// mBulletSize*20;
        this.height = mBulletSize;
        this.tall = mBulletSize;
        this.angle = obj.player.angle;
        this.angle2 = obj.player.angle2;
        this.player = obj.player;
        this.damageV = 20;
        this.isSquat = 0;
        this.dx = 0.0;
        this.dy = 0.0;
        this.dz = 0.0;
    }
    remove(){
        delete this.player.bullets[this.id];
        delete bullets[this.id];
    }
}

class BotPlayer extends Player{
    constructor(obj){
        super(obj);
        this.timer = setInterval(() => {
            this.timeCount+=1;
            this.isMoving=1;
            this.mode = 1;

            this.movement.forward = false;
            if(Math.random()*10<5){
                this.movement.forward = true;
            }
            
            if(! this.move(mPlayerSpeed*0.01)){
                this.angle = Math.random() * Math.PI * 2;
            }
            

            Object.values(players).forEach((player) => {
                if(player.isBot == 0){
                    var dx_ = player.x - this.x;
                    var dy_ = player.y - this.y;
                    var dz_ = player.z - this.z;
                    this.angle = Math.atan2(dy_,dx_);
                    var dr_ = Math.sqrt(dx_*dx_+dy_*dy_);
                    this.angle2 = -Math.atan2(dz_,dr_);
                    //console.log('angle2:'+ this.angle2/Math.PI*180.0+', ('+dz_+','+dr_);
                    return;
                }
             });


            if(Math.random()<0.01){
                this.shoot();
                var a_ = [1, 0.2];
            }
            if(Math.random()<0.01){
                this.squat();
            }
            if(Math.random()<0.01){
                this.jump();
            }
            if(Math.random()<0.01){
                this.angle = Math.random() * Math.PI * 2;
            }

    
        }, 1000/30);
    }
    remove(){
        const bandage = new Bandage({
            x: this.x,
            y: this.y,
            z: this.z,
        });
        bandages[bandage.id] = bandage;

        //if(Math.random()*10<5){
            const potion = new Potion({
                x: this.x + this.width,
                y: this.y + this.width,
                z: this.z,
            });
            potions[potion.id] = potion;    
        //}

        super.remove();
        clearInterval(this.timer);
        //reborn
        // setTimeout(() => {
        //     const bot = new BotPlayer({nickname: this.nickname});
        //     players[bot.id] = bot;
        // }, 3000);
    }
}



let players = {};
let bullets = {};
let walls = {};
let floors = {};
let slopes = {};
let cones = {};
let triangles = {};
let rocks = {};
let bandages = {};
let potions = {};
var mGlobalTime = 0;


function mCreateBandage(n_){
    for(let i=0; i<n_; i++){
        const bandage = new Bandage({
                x: Math.random() * FIELD_WIDTH,
                y: Math.random() * FIELD_HEIGHT,
                z: 0,
                angle: 0,
                angle2: 0,
                isSquat: 0,
        });
        bandages[bandage.id] = bandage;
    }
    
}

function mCreatePotion(n_){
    for(let i=0; i<n_; i++){
        const potion = new Potion({
                x: Math.random() * FIELD_WIDTH,
                y: Math.random() * FIELD_HEIGHT,
                z: 0,
                angle: 0,
                angle2: 0,
                isSquat: 0,
        });
        potions[potion.id] = potion;
    }    
}


function mCreateBot(n_){
    for(let i=0; i<n_; i++){
        const bot = new BotPlayer({nickname: 'bot'+i});
        bot.x = Math.random() * FIELD_WIDTH;
        bot.y = Math.random() * FIELD_WIDTH;
        bot.z = START_TALL*0.1;
        bot.isBot = 1;
        bot.socketId = 'bot';
        players[bot.id] = bot;
    }    
}


var mPlayerNum = 0;
var mBattleMode = 0;

io.on('connection', function(socket) {
    let player = null;
    socket.on('game-start', (config) => {
        player = new Player({
            socketId: socket.id,
            nickname: config.nickname,
        });
        player.mode = 0;
        player.timeCount = 0;
        player.order = mPlayerNum;
        players[player.id] = player;
        mPlayerNum += 1;
        io.to(player.socketId).emit('id', player.socketId);
        player.x = 100;
        player.y = 100;

        if(Object.keys(players).length == 1){
            if(mBattleMode==1){
                mCreateBot(9);
            }
            
            mGlobalTime = 0;
            mVictoryTime = 0;
            //mSetStorm();
        }
        // if(Object.keys(walls).length < 20){
        //     mCreateWall(20-Object.keys(walls).length);
        // }
        if(Object.keys(bandages).length < 20){
            mCreateBandage(20-Object.keys(bandages).length);
        }
        if(Object.keys(potions).length < 20){
            mCreatePotion(20-Object.keys(potions).length);
        }

    });
    socket.on('movement', function(movement) {
        if(!player || player.health===0){return;}
        player.movement = movement;
    });
    socket.on('mousemoveX', function(v) {
        if(!player || player.health===0){return;}
        //player.angle += v;
        player.angle -= v;
    });
    socket.on('mousemoveY', function(v) {
        if(!player || player.health===0){return;}
        player.angle2 += v;
        player.angle2 = Math.max(player.angle2,-3.1415/2*0.99);
        player.angle2 = Math.min(player.angle2,3.1415/2*0.99);
        //console.log('angle2:'+player.angle2);
    });
    // socket.on('shoot', function(){
    //     if(!player || player.health===0){return;}
    //     player.shoot();
    //     socket.broadcast.emit('shootFromOthers',0.3);
    // });
    socket.on('squat', function(){
        if(!player || player.health===0){return;}
        player.squat();
    });
    socket.on('jump', function(){
        if(!player || player.health===0){return;}
        player.jump();
    });
    socket.on('buildWall', function(){
        if(!player || player.health===0){return;}
        if(player.mode!=0){
            player.buildWall();
        }
    });
    socket.on('buildFloor', function(){
        if(!player || player.health===0){return;}
        if(player.mode!=0){
            player.buildFloor();
        }
    });
    socket.on('buildSlope', function(){
        console.log('buildSlope');
        if(!player || player.health===0){return;}
        if(player.mode!=0){
            player.buildSlope();
        }
    });
    socket.on('buildCone', function(){
        console.log('buildCone');
        if(!player || player.health===0){return;}
        if(player.mode!=0){
            player.buildCone();
        }
    });
    socket.on('doBuild', function(){
        if(!player || player.health===0){return;}
        if(player.mode!=0){
            player.doBuild();
        }
    });
    socket.on('jumper', function(){
        if(!player || player.health===0){return;}
        player.z = START_TALL;
        player.mode = 0;
    });
    socket.on('weaponChange', function(v_){
        if(!player || player.health===0){return;}
        player.mode = 1;
        player.weaponChange(v_);
    });
    socket.on('reload', function(){
        if(!player || player.health===0){return;}
        player.shootCount[player.weapon-1]=0;
    });
    socket.on('useBandage', function(){
        if(!player || player.health===0){return;}
        player.useBandage();
    });
    socket.on('usePotion', function(){
        if(!player || player.health===0){return;}
        player.usePotion();
    });
    socket.on('emote', function(){
        if(!player || player.health===0){return;}
        player.emote();
    });
    socket.on('modeChanged', function(v_){
        if(!player || player.health===0){return;}
        player.mode = v_;
    });
    socket.on('nowTouboBuild', function(v_){
        if(!player || player.health===0){return;}
        player.nowTouboBuild = v_;
        console.log('nowTouboBuild:'+player.nowTouboBuild);
    });
    socket.on('edit', function(){
        if(!player || player.health===0){return;}
        player.judgeEditMode();
    });
    socket.on('startSelectGrid', function(){
        if(!player || player.health===0){return;}
        console.log('startSelectGrid:'+player.buildType);
        player.isSelectingGrid = 1;
        if(player.buildType==0){
            player.selectGridWall();
        }
        else if(player.buildType==1){
            player.selectGridFloor();
        }
        else if(player.buildType==2){
            player.selectGridSlope();
        }
        else if(player.buildType==3){
            player.selectGridCone();
        }
    });
    socket.on('nowSelectGrid', function(){
        if(!player || player.health===0){return;}
        //console.log('nowSelectGrid:');
        if(player.buildType==0){
            player.selectGridWall();
        }
        else if(player.buildType==1){
            player.selectGridFloor();
        }
        else if(player.buildType==2){
            player.selectGridSlope();
        }
        else if(player.buildType==3){
            player.selectGridCone();
        }
    });
    socket.on('finishSelectGrid', function(){
        if(!player || player.health===0){return;}
        console.log('finishSelectGrid:');
        player.isSelectingGrid = 0;
        player.judgeEditMode();
    });
    socket.on('resetSelectGrid', function(){
        if(!player || player.health===0){return;}
        //console.log('nowSelectGrid:');
        player.resetSelectGrid();
    });
    socket.on('skinIsLoaded', function(){
        if(!player || player.health===0){return;}
        player.isSkinLoaded = 1;
    });


    socket.on('disconnect', () => {
        if(!player){return;}
        delete players[player.id];
        player = null;
        mPlayerNum -= 1;
        mPlayerNum = Math.max(0,mPlayerNum);
    });
});


function mRemoveAllPlayer()
{
    Object.values(players).forEach((player) => {
        player.remove();
    });
}

function mRemoveAllBuild(){
    Object.values(walls).forEach((wall) => {
        wall.remove();
    });
    Object.values(floors).forEach((floor) => {
        floor.remove();
    });
    Object.values(slopes).forEach((slope) => {
        slope.remove();
    });
}



var mDT = 0.1;
var mK = 80;
//var mC = 1;
var mC = 2 * Math.sqrt(mK) * 0.1 * 4;
mC = Math.min( mC  , 1.0/mDT *0.99   );
var mGravity = 9.8;
var mJumpVelocity = Math.sqrt( 2*mGravity*mJumpHight );
var mVictoryTime = 0;

setInterval(() => {

    const t_start = performance.now();
    
    mGlobalTime += 1;

    //Victory
    if(mBattleMode==1){
        if(Object.keys(players).length == 1){
            Object.values(players).forEach((player) => {
                if(player.isBot==0){
                    if(mVictoryTime == 0){
                        io.to(player.socketId).emit('victory');
                    }
                    else if(mVictoryTime >= GAME_FPS * 10){
                        mVictoryTime = 0;
                        player.remove();
                        mRemoveAllBuild();
                    }
                    mVictoryTime += 1;    
                }//

            });
            
        }
    }//


    Object.values(players).forEach((player) => {

        //var v_vec = [0,0,0];

        const movement = player.movement;
        const oldX = player.x, oldY = player.y, oldZ = player.z;
        player.oldX = oldX;
        player.oldY = oldY;
        player.oldZ = oldZ;

        player.timeCount += 1;
        //player.isMoving = 0;

        var s_ = mPlayerSpeed;
        if(player.isSquat==1){
            s_ = mPlayerSpeed / 2;
        }
        if(player.mode==0){
            s_ = mPlayerSpeed * 5;
        }

        var flag_jump = 0;
        if((player.isJump==1)  && (player.mode!=0)){
            player.velocity[2] += mJumpVelocity; //Math.sqrt( 2*mGravity*mJumpHight );
            player.isJump=0;
            flag_jump = 1;
        }


        if(!movement.forward){
            player.isMoving = 0;
        }

        if(movement.forward){
            player.move(s_);
        }
        if(movement.back){
            player.move(-s_);
        }
        if(movement.angup){
            player.angle2 -= mPlayerTurnSpeed/2;
        }
        if(movement.angdown){
            player.angle2 += mPlayerTurnSpeed/2;
        }
        if(movement.sideleft){
            player.side(s_);
        }
        if(movement.sideright){
            player.side(-s_);
        }

        if( player.angle > 2*3.1415 ){
            player.angle = player.angle - 2*3.1415;
        }
        if( player.angle < 0 ){
            player.angle = player.angle + 2*3.1415;
        }

        //console.log(' player:'+[player.x, player.y, player.z, Math.floor(player.angle/3.1415*180.0)] );


        //--- Physics
        // dt , x += v * dt, v += a * dt
        //var mDT = 0.1;

        // Gravity
        player.velocity[2] -= mGravity * mDT;

        // Velocity damping
        // player.velocity[0] -= player.velocity[0] * mC*0.001 * mDT;
        // player.velocity[1] -= player.velocity[1] * mC*0.001 * mDT;
        // player.velocity[2] -= player.velocity[2] * mC*0.001 * mDT;
        if(player.isGrounded==1){
            player.velocity[0] -= player.velocity[0] * 0.5 * mDT;
            player.velocity[1] -= player.velocity[1] * 0.5 * mDT;
            //player.velocity[2] -= player.velocity[2] * 0.1 * mDT;    
        }

        player.x += player.velocity[0] * mDT;
        player.y += player.velocity[1] * mDT;
        player.z += player.velocity[2] * mDT;

        //console.log('player.z:'+[player.z, player.velocity[2]] );
        //console.log('player:'+[player.x, player.y, player.z] );
        //console.log('player v:'+player.velocity);

        // build 
        if(player.mode == 2){
            if(player.buildType==0){
                player.buildWall();
            }
            else if(player.buildType==1){
                player.buildFloor();
            }            
            else if(player.buildType==2){
                player.buildSlope();
            }            
            else if(player.buildType==3){
                player.buildCone();
            }            
        }

        player.isShoot = 0;
        if(player.mode == 1)
        {
            if(movement.shoot){
                var mwp = player.weapon -1;
                if(player.timeCount>=player.shootTime+mShootCycle[mwp]) 
                {
                    if(player.shootCount[mwp]<mMagagineSize[mwp]){
                        player.isShoot = 1;
                        player.shoot();
                        player.firingTime = player.timeCount;
                        player.shootCount[mwp] += 1;    
                    }
                    else{                
                        io.to(player.socketId).emit('empty');
                    }
                    player.shootTime = player.timeCount;
                }
            }
        }//

        if((player.giveDamageTime+30>=player.timeCount) &&
           (player.giveDamage>0) ) 
        {
            io.to(player.socketId).emit('displayGiveDamage', player.timeCount-player.giveDamageTime);//1, player.giveDamage);
        }
        else{
            player.giveDamage = 0;
            io.to(player.socketId).emit('displayGiveDamage',0);
        }

        //var contactZ =0;
        player.isGrounded = 0;
   

        //Camera position
        player.cameraPosition[0] = -150;
        let cameraRay_ = player.getRayOfEye();
        cameraRay_.dx *= -1.0;
        cameraRay_.dy *= -1.0;
        cameraRay_.dz *= -1.0;
        ///console.log(' cameraRay_:'+[cameraRay_.x, cameraRay_.y, cameraRay_.z]);
        var dis_min = FIELD_WIDTH;
        var dis_lim = 2;

        var cc = 0;
        var cc_zp = 0;
        var cc_zm = 0;
        var ee = 0;
        var zz = [];


        //Triangle - Player Contact
        Object.values(triangles).forEach((triangle) => {

            //console.log('triangles['+triangle.id+']');

            if(triangle.isEditMode<2){

                //console.log('triangle face:'+);
                let array_i = [0];
                if(triangle.flag_backFace==1){
                    array_i = [0,1];
                }

                //var res_ = mPlanePointCollisionDepth(triangle, [0,1], player);
                var res_ = mPlanePointCollisionDepth(triangle, array_i, player);
                if(res_.judge){
                    console.log('triangle.parent:'+triangle.parentType);
                    console.log('triangle face array:'+array_i);
    
                    var vkx = res_.e * res_.nx * mK * mDT;
                    var vky = res_.e * res_.ny * mK * mDT;
                    var vkz = res_.e * res_.nz * mK * mDT;
                    //console.log('vk:'+[vkx, vky, vkz]);

                    var vcx = -player.velocity[0] * mC * mDT;
                    var vcy = -player.velocity[1] * mC * mDT;
                    var vcz = -player.velocity[2] * mC * mDT;
                    //console.log('vc:'+[vcx, vcy, vcz]);

                    player.velocity[0] += vkx + vcx;
                    player.velocity[1] += vky + vcy;
                    player.velocity[2] += vkz + vcz;

                    player.isGrounded = res_.isGrounded; // 1;
                    cc += 1;
                    if(res_.nz>0){
                        cc_zp += 1;
                    }
                    else if(res_.nz<0){
                        cc_zm += 1;
                    }    
                }// res
                else{
                    //Edge contact
                    res_ = mPlaneEdgePlayerCollisionDepth(triangle, 0, player);
                    if(res_.judge){
                        console.log('triangle face array:'+array_i);
                        console.log('triangle edge:'+res_.judge);
                        ee += 1;
                    }
               }


                var v_ = mGetRayCross(cameraRay_, triangle);
                if( (v_.judge) && (v_.e<dis_min)){
                    dis_min = v_.e;
                    //console.log(' camera distance:'+v_.e);
                    if(dis_min < 150){
                        player.cameraPosition[0] = - Math.max(dis_lim,dis_min);
                        //console.log(' camera distance:'+v_.e);
                    }//
                }//

            }//

        });


        // if(flag_jump==1){
        //     console.log('flag_jump:'+ player.velocity[2]);
        // }//

        
        //console.log('contact count:'+ cc);
        // console.log('         oldZ:'+ player.oldZ);
        // console.log('            z:'+ zz);

        //if( (cc>1) || (ee>0) ){
        if( (cc_zp>0) && (cc_zm>0) || (ee>0) ){
            console.log('contact count[zp,zm,edge]:'+ [cc_zp, cc_zm, ee]);
            // if(cc>1){
            //     console.log('contact count:'+ cc);
            // }//
            //player.x = player.oldX;
            //player.y = player.oldY;
            //player.z = (zz[0]+zz[1])/2;
            var dx = player.x - player.oldX;
            var dy = player.y - player.oldY;
            var dz = player.z - player.oldZ;

            player.x -= dx*1;
            player.y -= dy*1;
            //player.z -= dz*1;
        }//


        // Ground Contact
        if(player.z<0){
            //player.z = 0;
            player.isJump = 0;
            player.isGrounded = 1;

            // Contact force
            player.velocity[2] += (0-player.z) * mK * mDT;
            player.velocity[2] -= player.velocity[2] * mC * mDT;
            // Friction
            // if(Math.abs(player.velocity[0])>0){
            //     player.velocity[0] += -1.0 * Math.sign(player.velocity[0]) *  mGravity * mDT;
            // }
            // else if(Math.abs(player.velocity[1])>0){
            //     player.velocity[1] += -1.0 * Math.sign(player.velocity[1]) *  mGravity * mDT;
            // }
            
            if(player.mode == 0){
                player.mode = 1;
            }
        }

        // Avoid divergence
        if(player.velocity[2]>mJumpVelocity*1.5){
            player.velocity[2] *= 0.1;
        }

        if(player.x<0){
            player.x = 1.0;
        }
        if(player.y<0){
            player.y = 1.0;
        }
        if(player.x>FIELD_WIDTH){
            player.x = FIELD_WIDTH - 1.0;
        }
        if(player.y>FIELD_WIDTH){
            player.y = FIELD_WIDTH - 1.0;
        }


         Object.values(bandages).forEach((bandage) => {
            //if(player.intersect(bandage)){
            if(player.getItem(bandage)){
                player.bandages += 1;
                io.to(player.socketId).emit('getItem',1.0);
                bandage.remove();
            } 
         });

         Object.values(potions).forEach((potion) => {
            //if(player.intersect(potion)){
            if(player.getItem(potion)){
                player.potions += 1;
                io.to(player.socketId).emit('getItem',1.0);
                potion.remove();
            } 
         });

         //console.log("potions:"+Object.values(potions).length);
        

    }); //player

    //console.log("bullets:"+Object.values(bullets).length);
    Object.values(bullets).forEach((bullet) =>{

        //var c = 0;
        var e_min = mFieldGridSize*100;
        var id_ = 0;
        var type_ = -1;
        var hitPosition = [0,0,0];
        var hitTime = 0;
        

         Object.values(triangles).forEach((triangle) => {
            //console.log('bullet wall collision:'+bullet.id+','+wall.id);
            if(triangle.isEditMode!=2){
                for(var i=0;i<1;i++){
                    var res_ = mBulletWallCollision(triangle, i, bullet);
                    if( (res_.judge) && (res_.e<e_min) ){
                        id_ = triangle.id;
                        type_ = 3;
                        e_min = res_.e;
                        hitPosition = [res_.px,res_.py,res_.pz];
                    }//
                }
            }//
         });

         Object.values(players).forEach((player) => {

            var res_ = mBulletPlayerCollision(player, bullet);
            if( (res_.judge) && (res_.e<e_min) && player !== bullet.player ){
                id_ = player.id;
                type_ = 4;
                e_min = res_.e;
            }//

        });

        if(type_!=-1)
        {
            console.log('bullet:'+[e_min, id_]);

            if(type_==0){

            }
            else if(type_==1){

            }
            else if(type_==2){

            }
            else if(type_==3){
                let triangle = triangles[id_];
                if(triangle.parentType==0){
                    walls[triangle.parentID].damage(bullet.damageV);
                }
                else if(triangle.parentType==1){
                    floors[triangle.parentID].damage(bullet.damageV);
                }
                else if(triangle.parentType==2){
                    slopes[triangle.parentID].damage(bullet.damageV);
                }
                else if(triangle.parentType==3){
                    cones[triangle.parentID].damage(bullet.damageV);
                }
                triangle.hitPosition = hitPosition;
                triangle.hitTime = mGlobalTime;
            }
            else if(type_==4){
                let player = players[id_];
                player.damage(bullet.damageV);
                bullet.remove();
                bullet.player.point += 1;
                //console.log('bullet intersect player:'+bullet.id+','+player.id);

                let givePlayer = bullet.player;
                player.lastDamegePlayer = givePlayer.id;
                //io.to(bullet.player.socketId).emit('givedamage',1.0);
                io.to(givePlayer.socketId).emit('givedamage',1.0);

                bullet.player.giveDamage=bullet.damageV;
                bullet.player.giveDamageTime=bullet.player.timeCount;
            }
            
            bullet.remove();
            io.sockets.emit('walldamage', 0.5);
            return;           
        }

        bullet.remove();

    });

    //if(mGlobalTime%1==0){
        io.sockets.emit('state', mGlobalTime, players, bullets, walls, floors, slopes, cones, triangles, bandages, potions, mStorm);
    //}//
    

    const t_end = performance.now();
    var tcp = (t_end - t_start);
    // if(tcp>1){
    //     console.log('time:'+ (t_end - t_start) +'[ms]' );
    // }
    

}, 1000/GAME_FPS);


app.use('/static', express.static(__dirname + '/static'));

app.get('/', (request, response) => {
  response.sendFile(path.join(__dirname, '/static/index.html'));
  //response.sendFile(path.join(__dirname, '/static/3d.html'));
});

server.listen(PORT, function(){
    console.log('server listening. Port:' + PORT);
});
  